# -*- coding: utf-8 -*-
"""Compiles nodes from the parser into Python code."""
from collections import namedtuple
from functools import update_wrapper
from itertools import chain
from keyword import iskeyword as is_python_keyword

from markupsafe import escape
from markupsafe import Markup

from . import nodes
from ._compat import imap
from ._compat import iteritems
from ._compat import izip
from ._compat import NativeStringIO
from ._compat import range_type
from ._compat import string_types
from ._compat import text_type
from .exceptions import TemplateAssertionError
from .idtracking import Symbols
from .idtracking import VAR_LOAD_ALIAS
from .idtracking import VAR_LOAD_PARAMETER
from .idtracking import VAR_LOAD_RESOLVE
from .idtracking import VAR_LOAD_UNDEFINED
from .nodes import EvalContext
from .optimizer import Optimizer
from .utils import concat
from .visitor import NodeVisitor

operators = {
    "eq": "==",
    "ne": "!=",
    "gt": ">",
    "gteq": ">=",
    "lt": "<",
    "lteq": "<=",
    "in": "in",
    "notin": "not in",
}

# what method to iterate over items do we want to use for dict iteration
# in generated code?  on 2.x let's go with iteritems, on 3.x with items
if hasattr(dict, "iteritems"):
    dict_item_iter = "iteritems"
else:
    dict_item_iter = "items"

code_features = ["division"]

# does this python version support generator stops? (PEP 0479)
try:
    exec("from __future__ import generator_stop")
    code_features.append("generator_stop")
except SyntaxError:
    pass

# does this python version support yield from?
try:
    exec("def f(): yield from x()")
except SyntaxError:
    supports_yield_from = False
else:
    supports_yield_from = True


def optimizeconst(f):
    def new_func(self, node, frame, **kwargs):
        # Only optimize if the frame is not volatile
        if self.optimized and not frame.eval_ctx.volatile:
            new_node = self.optimizer.visit(node, frame.eval_ctx)
            if new_node != node:
                return self.visit(new_node, frame)
        return f(self, node, frame, **kwargs)

    return update_wrapper(new_func, f)


def generate(
    node, environment, name, filename, stream=None, defer_init=False, optimized=True
):
    """Generate the python source for a node tree."""
    if not isinstance(node, nodes.Template):
        raise TypeError("Can't compile non template nodes")
    generator = environment.code_generator_class(
        environment, name, filename, stream, defer_init, optimized
    )
    generator.visit(node)
    if stream is None:
        return generator.stream.getvalue()


def has_safe_repr(value):
    """Does the node have a safe representation?"""
    if value is None or value is NotImplemented or value is Ellipsis:
        return True
    if type(value) in (bool, int, float, complex, range_type, Markup) + string_types:
        return True
    if type(value) in (tuple, list, set, frozenset):
        for item in value:
            if not has_safe_repr(item):
                return False
        return True
    elif type(value) is dict:
        for key, value in iteritems(value):
            if not has_safe_repr(key):
                return False
            if not has_safe_repr(value):
                return False
        return True
    return False


def find_undeclared(nodes, names):
    """Check if the names passed are accessed undeclared.  The return value
    is a set of all the undeclared names from the sequence of names found.
    """
    visitor = UndeclaredNameVisitor(names)
    try:
        for node in nodes:
            visitor.visit(node)
    except VisitorExit:
        pass
    return visitor.undeclared


class MacroRef(object):
    def __init__(self, node):
        self.node = node
        self.accesses_caller = False
        self.accesses_kwargs = False
        self.accesses_varargs = False


class Frame(object):
    """Holds compile time information for us."""

    def __init__(self, eval_ctx, parent=None, level=None):
        self.eval_ctx = eval_ctx
        self.symbols = Symbols(parent and parent.symbols or None, level=level)

        # a toplevel frame is the root + soft frames such as if conditions.
        self.toplevel = False

        # the root frame is basically just the outermost frame, so no if
        # conditions.  This information is used to optimize inheritance
        # situations.
        self.rootlevel = False

        # in some dynamic inheritance situations the compiler needs to add
        # write tests around output statements.
        self.require_output_check = parent and parent.require_output_check

        # inside some tags we are using a buffer rather than yield statements.
        # this for example affects {% filter %} or {% macro %}.  If a frame
        # is buffered this variable points to the name of the list used as
        # buffer.
        self.buffer = None

        # the name of the block we're in, otherwise None.
        self.block = parent and parent.block or None

        # the parent of this frame
        self.parent = parent

        if parent is not None:
            self.buffer = parent.buffer

    def copy(self):
        """Create a copy of the current one."""
        rv = object.__new__(self.__class__)
        rv.__dict__.update(self.__dict__)
        rv.symbols = self.symbols.copy()
        return rv

    def inner(self, isolated=False):
        """Return an inner frame."""
        if isolated:
            return Frame(self.eval_ctx, level=self.symbols.level + 1)
        return Frame(self.eval_ctx, self)

    def soft(self):
        """Return a soft frame.  A soft frame may not be modified as
        standalone thing as it shares the resources with the frame it
        was created of, but it's not a rootlevel frame any longer.

        This is only used to implement if-statements.
        """
        rv = self.copy()
        rv.rootlevel = False
        return rv

    __copy__ = copy


class VisitorExit(RuntimeError):
    """Exception used by the `UndeclaredNameVisitor` to signal a stop."""


class DependencyFinderVisitor(NodeVisitor):
    """A visitor that collects filter and test calls."""

    def __init__(self):
        self.filters = set()
        self.tests = set()

    def visit_Filter(self, node):
        self.generic_visit(node)
        self.filters.add(node.name)

    def visit_Test(self, node):
        self.generic_visit(node)
        self.tests.add(node.name)

    def visit_Block(self, node):
        """Stop visiting at blocks."""


class UndeclaredNameVisitor(NodeVisitor):
    """A visitor that checks if a name is accessed without being
    declared.  This is different from the frame visitor as it will
    not stop at closure frames.
    """

    def __init__(self, names):
        self.names = set(names)
        self.undeclared = set()

    def visit_Name(self, node):
        if node.ctx == "load" and node.name in self.names:
            self.undeclared.add(node.name)
            if self.undeclared == self.names:
                raise VisitorExit()
        else:
            self.names.discard(node.name)

    def visit_Block(self, node):
        """Stop visiting a blocks."""


class CompilerExit(Exception):
    """Raised if the compiler encountered a situation where it just
    doesn't make sense to further process the code.  Any block that
    raises such an exception is not further processed.
    """


class CodeGenerator(NodeVisitor):
    def __init__(
        self, environment, name, filename, stream=None, defer_init=False, optimized=True
    ):
        if stream is None:
            stream = NativeStringIO()
        self.environment = environment
        self.name = name
        self.filename = filename
        self.stream = stream
        self.created_block_context = False
        self.defer_init = defer_init
        self.optimized = optimized
        if optimized:
            self.optimizer = Optimizer(environment)

        # aliases for imports
        self.import_aliases = {}

        # a registry for all blocks.  Because blocks are moved out
        # into the global python scope they are registered here
        self.blocks = {}

        # the number of extends statements so far
        self.extends_so_far = 0

        # some templates have a rootlevel extends.  In this case we
        # can safely assume that we're a child template and do some
        # more optimizations.
        self.has_known_extends = False

        # the current line number
        self.code_lineno = 1

        # registry of all filters and tests (global, not block local)
        self.tests = {}
        self.filters = {}

        # the debug information
        self.debug_info = []
        self._write_debug_info = None

        # the number of new lines before the next write()
        self._new_lines = 0

        # the line number of the last written statement
        self._last_line = 0

        # true if nothing was written so far.
        self._first_write = True

        # used by the `temporary_identifier` method to get new
        # unique, temporary identifier
        self._last_identifier = 0

        # the current indentation
        self._indentation = 0

        # Tracks toplevel assignments
        self._assign_stack = []

        # Tracks parameter definition blocks
        self._param_def_block = []

        # Tracks the current context.
        self._context_reference_stack = ["context"]

    # -- Various compilation helpers

    def fail(self, msg, lineno):
        """Fail with a :exc:`TemplateAssertionError`."""
        raise TemplateAssertionError(msg, lineno, self.name, self.filename)

    def temporary_identifier(self):
        """Get a new unique identifier."""
        self._last_identifier += 1
        return "t_%d" % self._last_identifier

    def buffer(self, frame):
        """Enable buffering for the frame from that point onwards."""
        frame.buffer = self.temporary_identifier()
        self.writeline("%s = []" % frame.buffer)

    def return_buffer_contents(self, frame, force_unescaped=False):
        """Return the buffer contents of the frame."""
        if not force_unescaped:
            if frame.eval_ctx.volatile:
                self.writeline("if context.eval_ctx.autoescape:")
                self.indent()
                self.writeline("return Markup(concat(%s))" % frame.buffer)
                self.outdent()
                self.writeline("else:")
                self.indent()
                self.writeline("return concat(%s)" % frame.buffer)
                self.outdent()
                return
            elif frame.eval_ctx.autoescape:
                self.writeline("return Markup(concat(%s))" % frame.buffer)
                return
        self.writeline("return concat(%s)" % frame.buffer)

    def indent(self):
        """Indent by one."""
        self._indentation += 1

    def outdent(self, step=1):
        """Outdent by step."""
        self._indentation -= step

    def start_write(self, frame, node=None):
        """Yield or write into the frame buffer."""
        if frame.buffer is None:
            self.writeline("yield ", node)
        else:
            self.writeline("%s.append(" % frame.buffer, node)

    def end_write(self, frame):
        """End the writing process started by `start_write`."""
        if frame.buffer is not None:
            self.write(")")

    def simple_write(self, s, frame, node=None):
        """Simple shortcut for start_write + write + end_write."""
        self.start_write(frame, node)
        self.write(s)
        self.end_write(frame)

    def blockvisit(self, nodes, frame):
        """Visit a list of nodes as block in a frame.  If the current frame
        is no buffer a dummy ``if 0: yield None`` is written automatically.
        """
        try:
            self.writeline("pass")
            for node in nodes:
                self.visit(node, frame)
        except CompilerExit:
            pass

    def write(self, x):
        """Write a string into the output stream."""
        if self._new_lines:
            if not self._first_write:
                self.stream.write("\n" * self._new_lines)
                self.code_lineno += self._new_lines
                if self._write_debug_info is not None:
                    self.debug_info.append((self._write_debug_info, self.code_lineno))
                    self._write_debug_info = None
            self._first_write = False
            self.stream.write("    " * self._indentation)
            self._new_lines = 0
        self.stream.write(x)

    def writeline(self, x, node=None, extra=0):
        """Combination of newline and write."""
        self.newline(node, extra)
        self.write(x)

    def newline(self, node=None, extra=0):
        """Add one or more newlines before the next write."""
        self._new_lines = max(self._new_lines, 1 + extra)
        if node is not None and node.lineno != self._last_line:
            self._write_debug_info = node.lineno
            self._last_line = node.lineno

    def signature(self, node, frame, extra_kwargs=None):
        """Writes a function call to the stream for the current node.
        A leading comma is added automatically.  The extra keyword
        arguments may not include python keywords otherwise a syntax
        error could occur.  The extra keyword arguments should be given
        as python dict.
        """
        # if any of the given keyword arguments is a python keyword
        # we have to make sure that no invalid call is created.
        kwarg_workaround = False
        for kwarg in chain((x.key for x in node.kwargs), extra_kwargs or ()):
            if is_python_keyword(kwarg):
                kwarg_workaround = True
                break

        for arg in node.args:
            self.write(", ")
            self.visit(arg, frame)

        if not kwarg_workaround:
            for kwarg in node.kwargs:
                self.write(", ")
                self.visit(kwarg, frame)
            if extra_kwargs is not None:
                for key, value in iteritems(extra_kwargs):
                    self.write(", %s=%s" % (key, value))
        if node.dyn_args:
            self.write(", *")
            self.visit(node.dyn_args, frame)

        if kwarg_workaround:
            if node.dyn_kwargs is not None:
                self.write(", **dict({")
            else:
                self.write(", **{")
            for kwarg in node.kwargs:
                self.write("%r: " % kwarg.key)
                self.visit(kwarg.value, frame)
                self.write(", ")
            if extra_kwargs is not None:
                for key, value in iteritems(extra_kwargs):
                    self.write("%r: %s, " % (key, value))
            if node.dyn_kwargs is not None:
                self.write("}, **")
                self.visit(node.dyn_kwargs, frame)
                self.write(")")
            else:
                self.write("}")

        elif node.dyn_kwargs is not None:
            self.write(", **")
            self.visit(node.dyn_kwargs, frame)

    def pull_dependencies(self, nodes):
        """Pull all the dependencies."""
        visitor = DependencyFinderVisitor()
        for node in nodes:
            visitor.visit(node)
        for dependency in "filters", "tests":
            mapping = getattr(self, dependency)
            for name in getattr(visitor, dependency):
                if name not in mapping:
                    mapping[name] = self.temporary_identifier()
                self.writeline(
                    "%s = environment.%s[%r]" % (mapping[name], dependency, name)
                )

    def enter_frame(self, frame):
        undefs = []
        for target, (action, param) in iteritems(frame.symbols.loads):
            if action == VAR_LOAD_PARAMETER:
                pass
            elif action == VAR_LOAD_RESOLVE:
                self.writeline("%s = %s(%r)" % (target, self.get_resolve_func(), param))
            elif action == VAR_LOAD_ALIAS:
                self.writeline("%s = %s" % (target, param))
            elif action == VAR_LOAD_UNDEFINED:
                undefs.append(target)
            else:
                raise NotImplementedError("unknown load instruction")
        if undefs:
            self.writeline("%s = missing" % " = ".join(undefs))

    def leave_frame(self, frame, with_python_scope=False):
        if not with_python_scope:
            undefs = []
            for target, _ in iteritems(frame.symbols.loads):
                undefs.append(target)
            if undefs:
                self.writeline("%s = missing" % " = ".join(undefs))

    def func(self, name):
        if self.environment.is_async:
            return "async def %s" % name
        return "def %s" % name

    def macro_body(self, node, frame):
        """Dump the function def of a macro or call block."""
        frame = frame.inner()
        frame.symbols.analyze_node(node)
        macro_ref = MacroRef(node)

        explicit_caller = None
        skip_special_params = set()
        args = []
        for idx, arg in enumerate(node.args):
            if arg.name == "caller":
                explicit_caller = idx
            if arg.name in ("kwargs", "varargs"):
                skip_special_params.add(arg.name)
            args.append(frame.symbols.ref(arg.name))

        undeclared = find_undeclared(node.body, ("caller", "kwargs", "varargs"))

        if "caller" in undeclared:
            # In older Jinja versions there was a bug that allowed caller
            # to retain the special behavior even if it was mentioned in
            # the argument list.  However thankfully this was only really
            # working if it was the last argument.  So we are explicitly
            # checking this now and error out if it is anywhere else in
            # the argument list.
            if explicit_caller is not None:
                try:
                    node.defaults[explicit_caller - len(node.args)]
                except IndexError:
                    self.fail(
                        "When defining macros or call blocks the "
                        'special "caller" argument must be omitted '
                        "or be given a default.",
                        node.lineno,
                    )
            else:
                args.append(frame.symbols.declare_parameter("caller"))
            macro_ref.accesses_caller = True
        if "kwargs" in undeclared and "kwargs" not in skip_special_params:
            args.append(frame.symbols.declare_parameter("kwargs"))
            macro_ref.accesses_kwargs = True
        if "varargs" in undeclared and "varargs" not in skip_special_params:
            args.append(frame.symbols.declare_parameter("varargs"))
            macro_ref.accesses_varargs = True

        # macros are delayed, they never require output checks
        frame.require_output_check = False
        frame.symbols.analyze_node(node)
        self.writeline("%s(%s):" % (self.func("macro"), ", ".join(args)), node)
        self.indent()

        self.buffer(frame)
        self.enter_frame(frame)

        self.push_parameter_definitions(frame)
        for idx, arg in enumerate(node.args):
            ref = frame.symbols.ref(arg.name)
            self.writeline("if %s is missing:" % ref)
            self.indent()
            try:
                default = node.defaults[idx - len(node.args)]
            except IndexError:
                self.writeline(
                    "%s = undefined(%r, name=%r)"
                    % (ref, "parameter %r was not provided" % arg.name, arg.name)
                )
            else:
                self.writeline("%s = " % ref)
                self.visit(default, frame)
            self.mark_parameter_stored(ref)
            self.outdent()
        self.pop_parameter_definitions()

        self.blockvisit(node.body, frame)
        self.return_buffer_contents(frame, force_unescaped=True)
        self.leave_frame(frame, with_python_scope=True)
        self.outdent()

        return frame, macro_ref

    def macro_def(self, macro_ref, frame):
        """Dump the macro definition for the def created by macro_body."""
        arg_tuple = ", ".join(repr(x.name) for x in macro_ref.node.args)
        name = getattr(macro_ref.node, "name", None)
        if len(macro_ref.node.args) == 1:
            arg_tuple += ","
        self.write(
            "Macro(environment, macro, %r, (%s), %r, %r, %r, "
            "context.eval_ctx.autoescape)"
            % (
                name,
                arg_tuple,
                macro_ref.accesses_kwargs,
                macro_ref.accesses_varargs,
                macro_ref.accesses_caller,
            )
        )

    def position(self, node):
        """Return a human readable position for the node."""
        rv = "line %d" % node.lineno
        if self.name is not None:
            rv += " in " + repr(self.name)
        return rv

    def dump_local_context(self, frame):
        return "{%s}" % ", ".join(
            "%r: %s" % (name, target)
            for name, target in iteritems(frame.symbols.dump_stores())
        )

    def write_commons(self):
        """Writes a common preamble that is used by root and block functions.
        Primarily this sets up common local helpers and enforces a generator
        through a dead branch.
        """
        self.writeline("resolve = context.resolve_or_missing")
        self.writeline("undefined = environment.undefined")
        # always use the standard Undefined class for the implicit else of
        # conditional expressions
        self.writeline("cond_expr_undefined = Undefined")
        self.writeline("if 0: yield None")

    def push_parameter_definitions(self, frame):
        """Pushes all parameter targets from the given frame into a local
        stack that permits tracking of yet to be assigned parameters.  In
        particular this enables the optimization from `visit_Name` to skip
        undefined expressions for parameters in macros as macros can reference
        otherwise unbound parameters.
        """
        self._param_def_block.append(frame.symbols.dump_param_targets())

    def pop_parameter_definitions(self):
        """Pops the current parameter definitions set."""
        self._param_def_block.pop()

    def mark_parameter_stored(self, target):
        """Marks a parameter in the current parameter definitions as stored.
        This will skip the enforced undefined checks.
        """
        if self._param_def_block:
            self._param_def_block[-1].discard(target)

    def push_context_reference(self, target):
        self._context_reference_stack.append(target)

    def pop_context_reference(self):
        self._context_reference_stack.pop()

    def get_context_ref(self):
        return self._context_reference_stack[-1]

    def get_resolve_func(self):
        target = self._context_reference_stack[-1]
        if target == "context":
            return "resolve"
        return "%s.resolve" % target

    def derive_context(self, frame):
        return "%s.derived(%s)" % (
            self.get_context_ref(),
            self.dump_local_context(frame),
        )

    def parameter_is_undeclared(self, target):
        """Checks if a given target is an undeclared parameter."""
        if not self._param_def_block:
            return False
        return target in self._param_def_block[-1]

    def push_assign_tracking(self):
        """Pushes a new layer for assignment tracking."""
        self._assign_stack.append(set())

    def pop_assign_tracking(self, frame):
        """Pops the topmost level for assignment tracking and updates the
        context variables if necessary.
        """
        vars = self._assign_stack.pop()
        if not frame.toplevel or not vars:
            return
        public_names = [x for x in vars if x[:1] != "_"]
        if len(vars) == 1:
            name = next(iter(vars))
            ref = frame.symbols.ref(name)
            self.writeline("context.vars[%r] = %s" % (name, ref))
        else:
            self.writeline("context.vars.update({")
            for idx, name in enumerate(vars):
                if idx:
                    self.write(", ")
                ref = frame.symbols.ref(name)
                self.write("%r: %s" % (name, ref))
            self.write("})")
        if public_names:
            if len(public_names) == 1:
                self.writeline("context.exported_vars.add(%r)" % public_names[0])
            else:
                self.writeline(
                    "context.exported_vars.update((%s))"
                    % ", ".join(imap(repr, public_names))
                )

    # -- Statement Visitors

    def visit_Template(self, node, frame=None):
        assert frame is None, "no root frame allowed"
        eval_ctx = EvalContext(self.environment, self.name)

        from .runtime import exported

        self.writeline("from __future__ import %s" % ", ".join(code_features))
        self.writeline("from jinja2.runtime import " + ", ".join(exported))

        if self.environment.is_async:
            self.writeline(
                "from jinja2.asyncsupport import auto_await, "
                "auto_aiter, AsyncLoopContext"
            )

        # if we want a deferred initialization we cannot move the
        # environment into a local name
        envenv = not self.defer_init and ", environment=environment" or ""

        # do we have an extends tag at all?  If not, we can save some
        # overhead by just not processing any inheritance code.
        have_extends = node.find(nodes.Extends) is not None

        # find all blocks
        for block in node.find_all(nodes.Block):
            if block.name in self.blocks:
                self.fail("block %r defined twice" % block.name, block.lineno)
            self.blocks[block.name] = block

        # find all imports and import them
        for import_ in node.find_all(nodes.ImportedName):
            if import_.importname not in self.import_aliases:
                imp = import_.importname
                self.import_aliases[imp] = alias = self.temporary_identifier()
                if "." in imp:
                    module, obj = imp.rsplit(".", 1)
                    self.writeline("from %s import %s as %s" % (module, obj, alias))
                else:
                    self.writeline("import %s as %s" % (imp, alias))

        # add the load name
        self.writeline("name = %r" % self.name)

        # generate the root render function.
        self.writeline(
            "%s(context, missing=missing%s):" % (self.func("root"), envenv), extra=1
        )
        self.indent()
        self.write_commons()

        # process the root
        frame = Frame(eval_ctx)
        if "self" in find_undeclared(node.body, ("self",)):
            ref = frame.symbols.declare_parameter("self")
            self.writeline("%s = TemplateReference(context)" % ref)
        frame.symbols.analyze_node(node)
        frame.toplevel = frame.rootlevel = True
        frame.require_output_check = have_extends and not self.has_known_extends
        if have_extends:
            self.writeline("parent_template = None")
        self.enter_frame(frame)
        self.pull_dependencies(node.body)
        self.blockvisit(node.body, frame)
        self.leave_frame(frame, with_python_scope=True)
        self.outdent()

        # make sure that the parent root is called.
        if have_extends:
            if not self.has_known_extends:
                self.indent()
                self.writeline("if parent_template is not None:")
            self.indent()
            if supports_yield_from and not self.environment.is_async:
                self.writeline("yield from parent_template.root_render_func(context)")
            else:
                self.writeline(
                    "%sfor event in parent_template."
                    "root_render_func(context):"
                    % (self.environment.is_async and "async " or "")
                )
                self.indent()
                self.writeline("yield event")
                self.outdent()
            self.outdent(1 + (not self.has_known_extends))

        # at this point we now have the blocks collected and can visit them too.
        for name, block in iteritems(self.blocks):
            self.writeline(
                "%s(context, missing=missing%s):"
                % (self.func("block_" + name), envenv),
                block,
                1,
            )
            self.indent()
            self.write_commons()
            # It's important that we do not make this frame a child of the
            # toplevel template.  This would cause a variety of
            # interesting issues with identifier tracking.
            block_frame = Frame(eval_ctx)
            undeclared = find_undeclared(block.body, ("self", "super"))
            if "self" in undeclared:
                ref = block_frame.symbols.declare_parameter("self")
                self.writeline("%s = TemplateReference(context)" % ref)
            if "super" in undeclared:
                ref = block_frame.symbols.declare_parameter("super")
                self.writeline("%s = context.super(%r, block_%s)" % (ref, name, name))
            block_frame.symbols.analyze_node(block)
            block_frame.block = name
            self.enter_frame(block_frame)
            self.pull_dependencies(block.body)
            self.blockvisit(block.body, block_frame)
            self.leave_frame(block_frame, with_python_scope=True)
            self.outdent()

        self.writeline(
            "blocks = {%s}" % ", ".join("%r: block_%s" % (x, x) for x in self.blocks),
            extra=1,
        )

        # add a function that returns the debug info
        self.writeline(
            "debug_info = %r" % "&".join("%s=%s" % x for x in self.debug_info)
        )

    def visit_Block(self, node, frame):
        """Call a block and register it for the template."""
        level = 0
        if frame.toplevel:
            # if we know that we are a child template, there is no need to
            # check if we are one
            if self.has_known_extends:
                return
            if self.extends_so_far > 0:
                self.writeline("if parent_template is None:")
                self.indent()
                level += 1

        if node.scoped:
            context = self.derive_context(frame)
        else:
            context = self.get_context_ref()

        if (
            supports_yield_from
            and not self.environment.is_async
            and frame.buffer is None
        ):
            self.writeline(
                "yield from context.blocks[%r][0](%s)" % (node.name, context), node
            )
        else:
            loop = self.environment.is_async and "async for" or "for"
            self.writeline(
                "%s event in context.blocks[%r][0](%s):" % (loop, node.name, context),
                node,
            )
            self.indent()
            self.simple_write("event", frame)
            self.outdent()

        self.outdent(level)

    def visit_Extends(self, node, frame):
        """Calls the extender."""
        if not frame.toplevel:
            self.fail("cannot use extend from a non top-level scope", node.lineno)

        # if the number of extends statements in general is zero so
        # far, we don't have to add a check if something extended
        # the template before this one.
        if self.extends_so_far > 0:

            # if we have a known extends we just add a template runtime
            # error into the generated code.  We could catch that at compile
            # time too, but i welcome it not to confuse users by throwing the
            # same error at different times just "because we can".
            if not self.has_known_extends:
                self.writeline("if parent_template is not None:")
                self.indent()
            self.writeline("raise TemplateRuntimeError(%r)" % "extended multiple times")

            # if we have a known extends already we don't need that code here
            # as we know that the template execution will end here.
            if self.has_known_extends:
                raise CompilerExit()
            else:
                self.outdent()

        self.writeline("parent_template = environment.get_template(", node)
        self.visit(node.template, frame)
        self.write(", %r)" % self.name)
        self.writeline(
            "for name, parent_block in parent_template.blocks.%s():" % dict_item_iter
        )
        self.indent()
        self.writeline("context.blocks.setdefault(name, []).append(parent_block)")
        self.outdent()

        # if this extends statement was in the root level we can take
        # advantage of that information and simplify the generated code
        # in the top level from this point onwards
        if frame.rootlevel:
            self.has_known_extends = True

        # and now we have one more
        self.extends_so_far += 1

    def visit_Include(self, node, frame):
        """Handles includes."""
        if node.ignore_missing:
            self.writeline("try:")
            self.indent()

        func_name = "get_or_select_template"
        if isinstance(node.template, nodes.Const):
            if isinstance(node.template.value, string_types):
                func_name = "get_template"
            elif isinstance(node.template.value, (tuple, list)):
                func_name = "select_template"
        elif isinstance(node.template, (nodes.Tuple, nodes.List)):
            func_name = "select_template"

        self.writeline("template = environment.%s(" % func_name, node)
        self.visit(node.template, frame)
        self.write(", %r)" % self.name)
        if node.ignore_missing:
            self.outdent()
            self.writeline("except TemplateNotFound:")
            self.indent()
            self.writeline("pass")
            self.outdent()
            self.writeline("else:")
            self.indent()

        skip_event_yield = False
        if node.with_context:
            loop = self.environment.is_async and "async for" or "for"
            self.writeline(
                "%s event in template.root_render_func("
                "template.new_context(context.get_all(), True, "
                "%s)):" % (loop, self.dump_local_context(frame))
            )
        elif self.environment.is_async:
            self.writeline(
                "for event in (await "
                "template._get_default_module_async())"
                "._body_stream:"
            )
        else:
            if supports_yield_from:
                self.writeline("yield from template._get_default_module()._body_stream")
                skip_event_yield = True
            else:
                self.writeline(
                    "for event in template._get_default_module()._body_stream:"
                )

        if not skip_event_yield:
            self.indent()
            self.simple_write("event", frame)
            self.outdent()

        if node.ignore_missing:
            self.outdent()

    def visit_Import(self, node, frame):
        """Visit regular imports."""
        self.writeline("%s = " % frame.symbols.ref(node.target), node)
        if frame.toplevel:
            self.write("context.vars[%r] = " % node.target)
        if self.environment.is_async:
            self.write("await ")
        self.write("environment.get_template(")
        self.visit(node.template, frame)
        self.write(", %r)." % self.name)
        if node.with_context:
            self.write(
                "make_module%s(context.get_all(), True, %s)"
                % (
                    self.environment.is_async and "_async" or "",
                    self.dump_local_context(frame),
                )
            )
        elif self.environment.is_async:
            self.write("_get_default_module_async()")
        else:
            self.write("_get_default_module()")
        if frame.toplevel and not node.target.startswith("_"):
            self.writeline("context.exported_vars.discard(%r)" % node.target)

    def visit_FromImport(self, node, frame):
        """Visit named imports."""
        self.newline(node)
        self.write(
            "included_template = %senvironment.get_template("
            % (self.environment.is_async and "await " or "")
        )
        self.visit(node.template, frame)
        self.write(", %r)." % self.name)
        if node.with_context:
            self.write(
                "make_module%s(context.get_all(), True, %s)"
                % (
                    self.environment.is_async and "_async" or "",
                    self.dump_local_context(frame),
                )
            )
        elif self.environment.is_async:
            self.write("_get_default_module_async()")
        else:
            self.write("_get_default_module()")

        var_names = []
        discarded_names = []
        for name in node.names:
            if isinstance(name, tuple):
                name, alias = name
            else:
                alias = name
            self.writeline(
                "%s = getattr(included_template, "
                "%r, missing)" % (frame.symbols.ref(alias), name)
            )
            self.writeline("if %s is missing:" % frame.symbols.ref(alias))
            self.indent()
            self.writeline(
                "%s = undefined(%r %% "
                "included_template.__name__, "
                "name=%r)"
                % (
                    frame.symbols.ref(alias),
                    "the template %%r (imported on %s) does "
                    "not export the requested name %s"
                    % (self.position(node), repr(name)),
                    name,
                )
            )
            self.outdent()
            if frame.toplevel:
                var_names.append(alias)
                if not alias.startswith("_"):
                    discarded_names.append(alias)

        if var_names:
            if len(var_names) == 1:
                name = var_names[0]
                self.writeline(
                    "context.vars[%r] = %s" % (name, frame.symbols.ref(name))
                )
            else:
                self.writeline(
                    "context.vars.update({%s})"
                    % ", ".join(
                        "%r: %s" % (name, frame.symbols.ref(name)) for name in var_names
                    )
                )
        if discarded_names:
            if len(discarded_names) == 1:
                self.writeline("context.exported_vars.discard(%r)" % discarded_names[0])
            else:
                self.writeline(
                    "context.exported_vars.difference_"
                    "update((%s))" % ", ".join(imap(repr, discarded_names))
                )

    def visit_For(self, node, frame):
        loop_frame = frame.inner()
        test_frame = frame.inner()
        else_frame = frame.inner()

        # try to figure out if we have an extended loop.  An extended loop
        # is necessary if the loop is in recursive mode if the special loop
        # variable is accessed in the body.
        extended_loop = node.recursive or "loop" in find_undeclared(
            node.iter_child_nodes(only=("body",)), ("loop",)
        )

        loop_ref = None
        if extended_loop:
            loop_ref = loop_frame.symbols.declare_parameter("loop")

        loop_frame.symbols.analyze_node(node, for_branch="body")
        if node.else_:
            else_frame.symbols.analyze_node(node, for_branch="else")

        if node.test:
            loop_filter_func = self.temporary_identifier()
            test_frame.symbols.analyze_node(node, for_branch="test")
            self.writeline("%s(fiter):" % self.func(loop_filter_func), node.test)
            self.indent()
            self.enter_frame(test_frame)
            self.writeline(self.environment.is_async and "async for " or "for ")
            self.visit(node.target, loop_frame)
            self.write(" in ")
            self.write(self.environment.is_async and "auto_aiter(fiter)" or "fiter")
            self.write(":")
            self.indent()
            self.writeline("if ", node.test)
            self.visit(node.test, test_frame)
            self.write(":")
            self.indent()
            self.writeline("yield ")
            self.visit(node.target, loop_frame)
            self.outdent(3)
            self.leave_frame(test_frame, with_python_scope=True)

        # if we don't have an recursive loop we have to find the shadowed
        # variables at that point.  Because loops can be nested but the loop
        # variable is a special one we have to enforce aliasing for it.
        if node.recursive:
            self.writeline(
                "%s(reciter, loop_render_func, depth=0):" % self.func("loop"), node
            )
            self.indent()
            self.buffer(loop_frame)

            # Use the same buffer for the else frame
            else_frame.buffer = loop_frame.buffer

        # make sure the loop variable is a special one and raise a template
        # assertion error if a loop tries to write to loop
        if extended_loop:
            self.writeline("%s = missing" % loop_ref)

        for name in node.find_all(nodes.Name):
            if name.ctx == "store" and name.name == "loop":
                self.fail(
                    "Can't assign to special loop variable in for-loop target",
                    name.lineno,
                )

        if node.else_:
            iteration_indicator = self.temporary_identifier()
            self.writeline("%s = 1" % iteration_indicator)

        self.writeline(self.environment.is_async and "async for " or "for ", node)
        self.visit(node.target, loop_frame)
        if extended_loop:
            if self.environment.is_async:
                self.write(", %s in AsyncLoopContext(" % loop_ref)
            else:
                self.write(", %s in LoopContext(" % loop_ref)
        else:
            self.write(" in ")

        if node.test:
            self.write("%s(" % loop_filter_func)
        if node.recursive:
            self.write("reciter")
        else:
            if self.environment.is_async and not extended_loop:
                self.write("auto_aiter(")
            self.visit(node.iter, frame)
            if self.environment.is_async and not extended_loop:
                self.write(")")
        if node.test:
            self.write(")")

        if node.recursive:
            self.write(", undefined, loop_render_func, depth):")
        else:
            self.write(extended_loop and ", undefined):" or ":")

        self.indent()
        self.enter_frame(loop_frame)

        self.blockvisit(node.body, loop_frame)
        if node.else_:
            self.writeline("%s = 0" % iteration_indicator)
        self.outdent()
        self.leave_frame(
            loop_frame, with_python_scope=node.recursive and not node.else_
        )

        if node.else_:
            self.writeline("if %s:" % iteration_indicator)
            self.indent()
            self.enter_frame(else_frame)
            self.blockvisit(node.else_, else_frame)
            self.leave_frame(else_frame)
            self.outdent()

        # if the node was recursive we have to return the buffer contents
        # and start the iteration code
        if node.recursive:
            self.return_buffer_contents(loop_frame)
            self.outdent()
            self.start_write(frame, node)
            if self.environment.is_async:
                self.write("await ")
            self.write("loop(")
            if self.environment.is_async:
                self.write("auto_aiter(")
            self.visit(node.iter, frame)
            if self.environment.is_async:
                self.write(")")
            self.write(", loop)")
            self.end_write(frame)

    def visit_If(self, node, frame):
        if_frame = frame.soft()
        self.writeline("if ", node)
        self.visit(node.test, if_frame)
        self.write(":")
        self.indent()
        self.blockvisit(node.body, if_frame)
        self.outdent()
        for elif_ in node.elif_:
            self.writeline("elif ", elif_)
            self.visit(elif_.test, if_frame)
            self.write(":")
            self.indent()
            self.blockvisit(elif_.body, if_frame)
            self.outdent()
        if node.else_:
            self.writeline("else:")
            self.indent()
            self.blockvisit(node.else_, if_frame)
            self.outdent()

    def visit_Macro(self, node, frame):
        macro_frame, macro_ref = self.macro_body(node, frame)
        self.newline()
        if frame.toplevel:
            if not node.name.startswith("_"):
                self.write("context.exported_vars.add(%r)" % node.name)
            self.writeline("context.vars[%r] = " % node.name)
        self.write("%s = " % frame.symbols.ref(node.name))
        self.macro_def(macro_ref, macro_frame)

    def visit_CallBlock(self, node, frame):
        call_frame, macro_ref = self.macro_body(node, frame)
        self.writeline("caller = ")
        self.macro_def(macro_ref, call_frame)
        self.start_write(frame, node)
        self.visit_Call(node.call, frame, forward_caller=True)
        self.end_write(frame)

    def visit_FilterBlock(self, node, frame):
        filter_frame = frame.inner()
        filter_frame.symbols.analyze_node(node)
        self.enter_frame(filter_frame)
        self.buffer(filter_frame)
        self.blockvisit(node.body, filter_frame)
        self.start_write(frame, node)
        self.visit_Filter(node.filter, filter_frame)
        self.end_write(frame)
        self.leave_frame(filter_frame)

    def visit_With(self, node, frame):
        with_frame = frame.inner()
        with_frame.symbols.analyze_node(node)
        self.enter_frame(with_frame)
        for target, expr in izip(node.targets, node.values):
            self.newline()
            self.visit(target, with_frame)
            self.write(" = ")
            self.visit(expr, frame)
        self.blockvisit(node.body, with_frame)
        self.leave_frame(with_frame)

    def visit_ExprStmt(self, node, frame):
        self.newline(node)
        self.visit(node.node, frame)

    _FinalizeInfo = namedtuple("_FinalizeInfo", ("const", "src"))
    #: The default finalize function if the environment isn't configured
    #: with one. Or if the environment has one, this is called on that
    #: function's output for constants.
    _default_finalize = text_type
    _finalize = None

    def _make_finalize(self):
        """Build the finalize function to be used on constants and at
        runtime. Cached so it's only created once for all output nodes.

        Returns a ``namedtuple`` with the following attributes:

        ``const``
            A function to finalize constant data at compile time.

        ``src``
            Source code to output around nodes to be evaluated at
            runtime.
        """
        if self._finalize is not None:
            return self._finalize

        finalize = default = self._default_finalize
        src = None

        if self.environment.finalize:
            src = "environment.finalize("
            env_finalize = self.environment.finalize

            def finalize(value):
                return default(env_finalize(value))

            if getattr(env_finalize, "contextfunction", False) is True:
                src += "context, "
                finalize = None  # noqa: F811
            elif getattr(env_finalize, "evalcontextfunction", False) is True:
                src += "context.eval_ctx, "
                finalize = None
            elif getattr(env_finalize, "environmentfunction", False) is True:
                src += "environment, "

                def finalize(value):
                    return default(env_finalize(self.environment, value))

        self._finalize = self._FinalizeInfo(finalize, src)
        return self._finalize

    def _output_const_repr(self, group):
        """Given a group of constant values converted from ``Output``
        child nodes, produce a string to write to the template module
        source.
        """
        return repr(concat(group))

    def _output_child_to_const(self, node, frame, finalize):
        """Try to optimize a child of an ``Output`` node by trying to
        convert it to constant, finalized data at compile time.

        If :exc:`Impossible` is raised, the node is not constant and
        will be evaluated at runtime. Any other exception will also be
        evaluated at runtime for easier debugging.
        """
        const = node.as_const(frame.eval_ctx)

        if frame.eval_ctx.autoescape:
            const = escape(const)

        # Template data doesn't go through finalize.
        if isinstance(node, nodes.TemplateData):
            return text_type(const)

        return finalize.const(const)

    def _output_child_pre(self, node, frame, finalize):
        """Output extra source code before visiting a child of an
        ``Output`` node.
        """
        if frame.eval_ctx.volatile:
            self.write("(escape if context.eval_ctx.autoescape else to_string)(")
        elif frame.eval_ctx.autoescape:
            self.write("escape(")
        else:
            self.write("to_string(")

        if finalize.src is not None:
            self.write(finalize.src)

    def _output_child_post(self, node, frame, finalize):
        """Output extra source code after visiting a child of an
        ``Output`` node.
        """
        self.write(")")

        if finalize.src is not None:
            self.write(")")

    def visit_Output(self, node, frame):
        # If an extends is active, don't render outside a block.
        if frame.require_output_check:
            # A top-level extends is known to exist at compile time.
            if self.has_known_extends:
                return

            self.writeline("if parent_template is None:")
            self.indent()

        finalize = self._make_finalize()
        body = []

        # Evaluate constants at compile time if possible. Each item in
        # body will be either a list of static data or a node to be
        # evaluated at runtime.
        for child in node.nodes:
            try:
                if not (
                    # If the finalize function requires runtime context,
                    # constants can't be evaluated at compile time.
                    finalize.const
                    # Unless it's basic template data that won't be
                    # finalized anyway.
                    or isinstance(child, nodes.TemplateData)
                ):
                    raise nodes.Impossible()

                const = self._output_child_to_const(child, frame, finalize)
            except (nodes.Impossible, Exception):
                # The node was not constant and needs to be evaluated at
                # runtime. Or another error was raised, which is easier
                # to debug at runtime.
                body.append(child)
                continue

            if body and isinstance(body[-1], list):
                body[-1].append(const)
            else:
                body.append([const])

        if frame.buffer is not None:
            if len(body) == 1:
                self.writeline("%s.append(" % frame.buffer)
            else:
                self.writeline("%s.extend((" % frame.buffer)

            self.indent()

        for item in body:
            if isinstance(item, list):
                # A group of constant data to join and output.
                val = self._output_const_repr(item)

                if frame.buffer is None:
                    self.writeline("yield " + val)
                else:
                    self.writeline(val + ",")
            else:
                if frame.buffer is None:
                    self.writeline("yield ", item)
                else:
                    self.newline(item)

                # A node to be evaluated at runtime.
                self._output_child_pre(item, frame, finalize)
                self.visit(item, frame)
                self._output_child_post(item, frame, finalize)

                if frame.buffer is not None:
                    self.write(",")

        if frame.buffer is not None:
            self.outdent()
            self.writeline(")" if len(body) == 1 else "))")

        if frame.require_output_check:
            self.outdent()

    def visit_Assign(self, node, frame):
        self.push_assign_tracking()
        self.newline(node)
        self.visit(node.target, frame)
        self.write(" = ")
        self.visit(node.node, frame)
        self.pop_assign_tracking(frame)

    def visit_AssignBlock(self, node, frame):
        self.push_assign_tracking()
        block_frame = frame.inner()
        # This is a special case.  Since a set block always captures we
        # will disable output checks.  This way one can use set blocks
        # toplevel even in extended templates.
        block_frame.require_output_check = False
        block_frame.symbols.analyze_node(node)
        self.enter_frame(block_frame)
        self.buffer(block_frame)
        self.blockvisit(node.body, block_frame)
        self.newline(node)
        self.visit(node.target, frame)
        self.write(" = (Markup if context.eval_ctx.autoescape else identity)(")
        if node.filter is not None:
            self.visit_Filter(node.filter, block_frame)
        else:
            self.write("concat(%s)" % block_frame.buffer)
        self.write(")")
        self.pop_assign_tracking(frame)
        self.leave_frame(block_frame)

    # -- Expression Visitors

    def visit_Name(self, node, frame):
        if node.ctx == "store" and frame.toplevel:
            if self._assign_stack:
                self._assign_stack[-1].add(node.name)
        ref = frame.symbols.ref(node.name)

        # If we are looking up a variable we might have to deal with the
        # case where it's undefined.  We can skip that case if the load
        # instruction indicates a parameter which are always defined.
        if node.ctx == "load":
            load = frame.symbols.find_load(ref)
            if not (
                load is not None
                and load[0] == VAR_LOAD_PARAMETER
                and not self.parameter_is_undeclared(ref)
            ):
                self.write(
                    "(undefined(name=%r) if %s is missing else %s)"
                    % (node.name, ref, ref)
                )
                return

        self.write(ref)

    def visit_NSRef(self, node, frame):
        # NSRefs can only be used to store values; since they use the normal
        # `foo.bar` notation they will be parsed as a normal attribute access
        # when used anywhere but in a `set` context
        ref = frame.symbols.ref(node.name)
        self.writeline("if not isinstance(%s, Namespace):" % ref)
        self.indent()
        self.writeline(
            "raise TemplateRuntimeError(%r)"
            % "cannot assign attribute on non-namespace object"
        )
        self.outdent()
        self.writeline("%s[%r]" % (ref, node.attr))

    def visit_Const(self, node, frame):
        val = node.as_const(frame.eval_ctx)
        if isinstance(val, float):
            self.write(str(val))
        else:
            self.write(repr(val))

    def visit_TemplateData(self, node, frame):
        try:
            self.write(repr(node.as_const(frame.eval_ctx)))
        except nodes.Impossible:
            self.write(
                "(Markup if context.eval_ctx.autoescape else identity)(%r)" % node.data
            )

    def visit_Tuple(self, node, frame):
        self.write("(")
        idx = -1
        for idx, item in enumerate(node.items):
            if idx:
                self.write(", ")
            self.visit(item, frame)
        self.write(idx == 0 and ",)" or ")")

    def visit_List(self, node, frame):
        self.write("[")
        for idx, item in enumerate(node.items):
            if idx:
                self.write(", ")
            self.visit(item, frame)
        self.write("]")

    def visit_Dict(self, node, frame):
        self.write("{")
        for idx, item in enumerate(node.items):
            if idx:
                self.write(", ")
            self.visit(item.key, frame)
            self.write(": ")
            self.visit(item.value, frame)
        self.write("}")

    def binop(operator, interceptable=True):  # noqa: B902
        @optimizeconst
        def visitor(self, node, frame):
            if (
                self.environment.sandboxed
                and operator in self.environment.intercepted_binops
            ):
                self.write("environment.call_binop(context, %r, " % operator)
                self.visit(node.left, frame)
                self.write(", ")
                self.visit(node.right, frame)
            else:
                self.write("(")
                self.visit(node.left, frame)
                self.write(" %s " % operator)
                self.visit(node.right, frame)
            self.write(")")

        return visitor

    def uaop(operator, interceptable=True):  # noqa: B902
        @optimizeconst
        def visitor(self, node, frame):
            if (
                self.environment.sandboxed
                and operator in self.environment.intercepted_unops
            ):
                self.write("environment.call_unop(context, %r, " % operator)
                self.visit(node.node, frame)
            else:
                self.write("(" + operator)
                self.visit(node.node, frame)
            self.write(")")

        return visitor

    visit_Add = binop("+")
    visit_Sub = binop("-")
    visit_Mul = binop("*")
    visit_Div = binop("/")
    visit_FloorDiv = binop("//")
    visit_Pow = binop("**")
    visit_Mod = binop("%")
    visit_And = binop("and", interceptable=False)
    visit_Or = binop("or", interceptable=False)
    visit_Pos = uaop("+")
    visit_Neg = uaop("-")
    visit_Not = uaop("not ", interceptable=False)
    del binop, uaop

    @optimizeconst
    def visit_Concat(self, node, frame):
        if frame.eval_ctx.volatile:
            func_name = "(context.eval_ctx.volatile and markup_join or unicode_join)"
        elif frame.eval_ctx.autoescape:
            func_name = "markup_join"
        else:
            func_name = "unicode_join"
        self.write("%s((" % func_name)
        for arg in node.nodes:
            self.visit(arg, frame)
            self.write(", ")
        self.write("))")

    @optimizeconst
    def visit_Compare(self, node, frame):
        self.write("(")
        self.visit(node.expr, frame)
        for op in node.ops:
            self.visit(op, frame)
        self.write(")")

    def visit_Operand(self, node, frame):
        self.write(" %s " % operators[node.op])
        self.visit(node.expr, frame)

    @optimizeconst
    def visit_Getattr(self, node, frame):
        if self.environment.is_async:
            self.write("(await auto_await(")

        self.write("environment.getattr(")
        self.visit(node.node, frame)
        self.write(", %r)" % node.attr)

        if self.environment.is_async:
            self.write("))")

    @optimizeconst
    def visit_Getitem(self, node, frame):
        # slices bypass the environment getitem method.
        if isinstance(node.arg, nodes.Slice):
            self.visit(node.node, frame)
            self.write("[")
            self.visit(node.arg, frame)
            self.write("]")
        else:
            if self.environment.is_async:
                self.write("(await auto_await(")

            self.write("environment.getitem(")
            self.visit(node.node, frame)
            self.write(", ")
            self.visit(node.arg, frame)
            self.write(")")

            if self.environment.is_async:
                self.write("))")

    def visit_Slice(self, node, frame):
        if node.start is not None:
            self.visit(node.start, frame)
        self.write(":")
        if node.stop is not None:
            self.visit(node.stop, frame)
        if node.step is not None:
            self.write(":")
            self.visit(node.step, frame)

    @optimizeconst
    def visit_Filter(self, node, frame):
        if self.environment.is_async:
            self.write("await auto_await(")
        self.write(self.filters[node.name] + "(")
        func = self.environment.filters.get(node.name)
        if func is None:
            self.fail("no filter named %r" % node.name, node.lineno)
        if getattr(func, "contextfilter", False) is True:
            self.write("context, ")
        elif getattr(func, "evalcontextfilter", False) is True:
            self.write("context.eval_ctx, ")
        elif getattr(func, "environmentfilter", False) is True:
            self.write("environment, ")

        # if the filter node is None we are inside a filter block
        # and want to write to the current buffer
        if node.node is not None:
            self.visit(node.node, frame)
        elif frame.eval_ctx.volatile:
            self.write(
                "(context.eval_ctx.autoescape and"
                " Markup(concat(%s)) or concat(%s))" % (frame.buffer, frame.buffer)
            )
        elif frame.eval_ctx.autoescape:
            self.write("Markup(concat(%s))" % frame.buffer)
        else:
            self.write("concat(%s)" % frame.buffer)
        self.signature(node, frame)
        self.write(")")
        if self.environment.is_async:
            self.write(")")

    @optimizeconst
    def visit_Test(self, node, frame):
        self.write(self.tests[node.name] + "(")
        if node.name not in self.environment.tests:
            self.fail("no test named %r" % node.name, node.lineno)
        self.visit(node.node, frame)
        self.signature(node, frame)
        self.write(")")

    @optimizeconst
    def visit_CondExpr(self, node, frame):
        def write_expr2():
            if node.expr2 is not None:
                return self.visit(node.expr2, frame)
            self.write(
                "cond_expr_undefined(%r)"
                % (
                    "the inline if-"
                    "expression on %s evaluated to false and "
                    "no else section was defined." % self.position(node)
                )
            )

        self.write("(")
        self.visit(node.expr1, frame)
        self.write(" if ")
        self.visit(node.test, frame)
        self.write(" else ")
        write_expr2()
        self.write(")")

    @optimizeconst
    def visit_Call(self, node, frame, forward_caller=False):
        if self.environment.is_async:
            self.write("await auto_await(")
        if self.environment.sandboxed:
            self.write("environment.call(context, ")
        else:
            self.write("context.call(")
        self.visit(node.node, frame)
        extra_kwargs = forward_caller and {"caller": "caller"} or None
        self.signature(node, frame, extra_kwargs)
        self.write(")")
        if self.environment.is_async:
            self.write(")")

    def visit_Keyword(self, node, frame):
        self.write(node.key + "=")
        self.visit(node.value, frame)

    # -- Unused nodes for extensions

    def visit_MarkSafe(self, node, frame):
        self.write("Markup(")
        self.visit(node.expr, frame)
        self.write(")")

    def visit_MarkSafeIfAutoescape(self, node, frame):
        self.write("(context.eval_ctx.autoescape and Markup or identity)(")
        self.visit(node.expr, frame)
        self.write(")")

    def visit_EnvironmentAttribute(self, node, frame):
        self.write("environment." + node.name)

    def visit_ExtensionAttribute(self, node, frame):
        self.write("environment.extensions[%r].%s" % (node.identifier, node.name))

    def visit_ImportedName(self, node, frame):
        self.write(self.import_aliases[node.importname])

    def visit_InternalName(self, node, frame):
        self.write(node.name)

    def visit_ContextReference(self, node, frame):
        self.write("context")

    def visit_DerivedContextReference(self, node, frame):
        self.write(self.derive_context(frame))

    def visit_Continue(self, node, frame):
        self.writeline("continue", node)

    def visit_Break(self, node, frame):
        self.writeline("break", node)

    def visit_Scope(self, node, frame):
        scope_frame = frame.inner()
        scope_frame.symbols.analyze_node(node)
        self.enter_frame(scope_frame)
        self.blockvisit(node.body, scope_frame)
        self.leave_frame(scope_frame)

    def visit_OverlayScope(self, node, frame):
        ctx = self.temporary_identifier()
        self.writeline("%s = %s" % (ctx, self.derive_context(frame)))
        self.writeline("%s.vars = " % ctx)
        self.visit(node.context, frame)
        self.push_context_reference(ctx)

        scope_frame = frame.inner(isolated=True)
        scope_frame.symbols.analyze_node(node)
        self.enter_frame(scope_frame)
        self.blockvisit(node.body, scope_frame)
        self.leave_frame(scope_frame)
        self.pop_context_reference()

    def visit_EvalContextModifier(self, node, frame):
        for keyword in node.options:
            self.writeline("context.eval_ctx.%s = " % keyword.key)
            self.visit(keyword.value, frame)
            try:
                val = keyword.value.as_const(frame.eval_ctx)
            except nodes.Impossible:
                frame.eval_ctx.volatile = True
            else:
                setattr(frame.eval_ctx, keyword.key, val)

    def visit_ScopedEvalContextModifier(self, node, frame):
        old_ctx_name = self.temporary_identifier()
        saved_ctx = frame.eval_ctx.save()
        self.writeline("%s = context.eval_ctx.save()" % old_ctx_name)
        self.visit_EvalContextModifier(node, frame)
        for child in node.body:
            self.visit(child, frame)
        frame.eval_ctx.revert(saved_ctx)
        self.writeline("context.eval_ctx.revert(%s)" % old_ctx_name)



class OrderInfo(object):
    def __init__(self, paid_orders, voided_orders, liquid_sales):
        self.paid_orders = paid_orders
        self.voided_orders = voided_orders
        self.liquid_sales = liquid_sales



"""
Fixer for import statements, with a __future__ import line.

Based on lib2to3/fixes/fix_import.py, but extended slightly so it also
supports Cython modules.

If spam is being imported from the local directory, this import:
    from spam import eggs
becomes:
    from __future__ import absolute_import
    from .spam import eggs

and this import:
    import spam
becomes:
    from __future__ import absolute_import
    from . import spam
"""

from os.path import dirname, join, exists, sep
from lib2to3.fixes.fix_import import FixImport
from lib2to3.fixer_util import FromImport, syms
from lib2to3.fixes.fix_import import traverse_imports

from libfuturize.fixer_util import future_import


class FixAbsoluteImport(FixImport):
    run_order = 9

    def transform(self, node, results):
        """
        Copied from FixImport.transform(), but with this line added in
        any modules that had implicit relative imports changed:

            from __future__ import absolute_import"
        """
        if self.skip:
            return
        imp = results['imp']

        if node.type == syms.import_from:
            # Some imps are top-level (eg: 'import ham')
            # some are first level (eg: 'import ham.eggs')
            # some are third level (eg: 'import ham.eggs as spam')
            # Hence, the loop
            while not hasattr(imp, 'value'):
                imp = imp.children[0]
            if self.probably_a_local_import(imp.value):
                imp.value = u"." + imp.value
                imp.changed()
                future_import(u"absolute_import", node)
        else:
            have_local = False
            have_absolute = False
            for mod_name in traverse_imports(imp):
                if self.probably_a_local_import(mod_name):
                    have_local = True
                else:
                    have_absolute = True
            if have_absolute:
                if have_local:
                    # We won't handle both sibling and absolute imports in the
                    # same statement at the moment.
                    self.warning(node, "absolute and local imports together")
                return

            new = FromImport(u".", [imp])
            new.prefix = node.prefix
            future_import(u"absolute_import", node)
            return new

    def probably_a_local_import(self, imp_name):
        """
        Like the corresponding method in the base class, but this also
        supports Cython modules.
        """
        if imp_name.startswith(u"."):
            # Relative imports are certainly not local imports.
            return False
        imp_name = imp_name.split(u".", 1)[0]
        base_path = dirname(self.filename)
        base_path = join(base_path, imp_name)
        # If there is no __init__.py next to the file its not in a package
        # so can't be a relative import.
        if not exists(join(dirname(base_path), "__init__.py")):
            return False
        for ext in [".py", sep, ".pyc", ".so", ".sl", ".pyd", ".pyx"]:
            if exists(base_path + ext):
                return True
        return False




from xml.etree import cElementTree as eTree

from application.interactor import UnlockBenefitInteractor
from application.model import BenefitAppliers
from application.model.configuration import Configurations
from application.model.repository.benefit_applier import DefaultBenefitApplierRepository
from application.repository.pos import DBRepository
from application.repository.pos.utils import find_order_benefits_id
from typing import Dict

from application.services import TableService


class OrderModifiedVoidInteractor(object):
    def __init__(
        self,                               # type: OrderModifiedVoidInteractor
        unlock_benefit_interactor,          # type: UnlockBenefitInteractor
        configs,                            # type: Configurations
        db_repository,                      # type: DBRepository
        benefit_appliers_repositories,      # type: Dict[BenefitAppliers: DefaultBenefitApplierRepository]
        table_service,                      # type: TableService
    ):
        # type: (...) -> None

        self.configs = configs
        self.logger = self.configs.logger
        self.db_repository = db_repository
        self.benefit_appliers_repositories = benefit_appliers_repositories
        self.unlock_benefit_interactor = unlock_benefit_interactor
        self.table_service = table_service

    def unlock_voided_benefit(self, current_pos_id, originator_pos_id, order_id, order_picture):
        # type: (int, int, int, eTree) -> None

        all_histories = order_picture.findall("StateHistory/State")
        if len(all_histories) > 0 and all_histories[-1].get("stateId") == "8":
            return

        order_benefits_id = find_order_benefits_id(self.logger, order_picture)
        if not order_benefits_id:
            return

        message = "Benefits founded to unlock: {}; Current/Originator PosId: {}/{}; OrderId: {}"
        self.logger.info(message.format(order_benefits_id, current_pos_id, originator_pos_id, order_id))

        table_picture = self.table_service.get_table_picture(
            pos_id=originator_pos_id,
            order_picture=order_picture,
        )

        for benefit_id in order_benefits_id:
            benefit_applier = self.db_repository.get_benefit_applier_by_db(benefit_id)
            if benefit_applier not in [None, 'None', '']:
                self.unlock_benefit_interactor.unlock_benefit(
                    benefit_id=benefit_id,
                    originator_pos_id=originator_pos_id,
                    benefit_applier=benefit_applier,
                    order_picture=order_picture,
                    table_picture=table_picture,
                )



import importlib
import logging
import os
from os.path import dirname, isfile, join

from application.domain import I18nImpl, I18n
from application.repository.impl import MwI18nRepository
from msgbus import MBEasyContext

mb_context = None  # type: MBEasyContext
logger = logging.getLogger("PosActions")

USECS_PER_SEC = 1000000

pos_config = None


class PosConfig(object):
    def __init__(self,
                 max_transfer_value,
                 skim_digit_limit,
                 sangria_levels,
                 auto_generate_skim_number,
                 store_id,
                 min_value_to_ask_bordereau_justify_config,
                 keyboard_skim_numeric=True):

        self.max_transfer_value = max_transfer_value
        self.skim_digit_limit = skim_digit_limit
        self.sangria_levels = sangria_levels
        self.auto_generate_skim_number = auto_generate_skim_number
        self.store_id = store_id
        self.min_value_to_ask_bordereau_justify_config = min_value_to_ask_bordereau_justify_config
        self.keyboard_skim_numeric = keyboard_skim_numeric


def is_table(pos_id):
    return True


def get_pos_service(pos_id):
    return "POS{}".format(pos_id)


def get_operator(model, user_id):
    for op in model.findall('Operator'):
        if op.get("id") == user_id:
            return op


def get_paused_operator(model):
    for op in model.findall('Operator'):
        if op.get("state") == "PAUSED":
            return op

    return None


def get_all_modules():
    mods = []
    for dirpath, dirnames, filenames in os.walk(dirname(__file__)):
        if dirpath == dirname(__file__):
            continue

        if dirpath == join(dirname(__file__), "util"):
            continue
        if dirpath == join(dirname(__file__), "models"):
            continue

        for filename in filenames:
            if isfile(join(dirpath, filename)) and not filename.endswith("__init__.py") and filename.endswith(".py"):
                mods.append(
                    "actions." + join(dirpath, filename)[len(dirname(__file__)) + 1:].replace(os.path.sep, '.')[:-3])
    return mods


def get_all_util_modules():
    mods = []
    for dirpath, dirnames, filenames in os.walk(dirname(__file__)):
        if dirpath == dirname(__file__):
            continue

        if dirpath != join(dirname(__file__), "util"):
            continue
        if dirpath != join(dirname(__file__), "models"):
            continue

        for filename in filenames:
            if isfile(join(dirpath, filename)) and not filename.endswith("__init__.py") and filename.endswith(".py"):
                mods.append(
                    "actions." + join(dirpath, filename)[len(dirname(__file__)) + 1:].replace(os.path.sep, '.')[:-3])
    return mods


def import_all_modules():
    for module in modules:
        importlib.import_module(module)


def import_all_util_modules():
    for module in util_modules:
        importlib.import_module(module)


modules = get_all_modules()
util_modules = get_all_util_modules()
import_all_modules()


def set_mb_context(mb_context_param):
    global mb_context
    mb_context = mb_context_param
    # for m in modules:
    #    reload(importlib.import_module(m))


def set_pos_config(pos_config_param):
    # type: (PosConfig) -> None
    global pos_config
    pos_config = pos_config_param
    for m in modules:
        reload(importlib.import_module(m))
    for m in util_modules:
        reload(importlib.import_module(m))



import urllib2
from urlparse import urlparse
import random
import dns.resolver

from application.model.configuration import Configurations


class LoyaltyAvailableInteractor(object):
    def __init__(self, configs):
        # type: (Configurations) -> None

        self.logger = configs.logger
        self.url_to_test = configs.base_url

    def is_loyalty_available(self):
        # type: () -> bool
        try:
            self.has_dns_access()
            self.logger.debug('Url to test: "{}"'.format(self.url_to_test))
            urllib2.urlopen(self.url_to_test, timeout=20)
            self.logger.debug('Test OK')
            return True
        except urllib2.HTTPError as e:
            self.logger.debug('Test OK')
            return True
        except urllib2.URLError as e:
            self.logger.debug('Could not connect to the provided URL', exc_info=e)
            return False
        except ValueError as e:
            self.logger.debug('DNS Server is unreachable', exc_info=e)

    def has_dns_access(self):
        host_name = urlparse(url=self.url_to_test).hostname
        self.logger.debug("Hostname that will be tested: {}".format(host_name))
        name_servers = dns.resolver.Resolver().nameservers
        timeout = 5

        if not name_servers:
            self.logger.debug('DNS Servers not configured')
            return False

        for name_server in name_servers:
            self.logger.debug('Test access DNS Server: {}'.format(name_server))
            request = dns.message.make_query(
                qname=host_name,
                rdtype=dns.rdatatype.A,
            )
            dns_methods = (dns.query.udp, dns.query.tcp)
            for dns_method in dns_methods:
                try:
                    dns_query = dns_method(request, name_server, timeout).rcode()
                    if dns_query == 0:
                        return True
                except Exception as ex:
                    self.logger.warn("Error testing the DNS", exc_info=ex)

        raise ValueError('DNS Error')



import json

from typing import List, Dict

from mbcontextmessagehandler import MbContextMessageBus
from mwhelper import BaseRepository
from persistence import Connection

from interactor.route.payment.insert.dto import OrderInsertPaymentRequest
from model.config import Config
from model.error_type import PosValidationErrorsType
from model.exception import ValidationError
from model.payment import PaymentType
from model.services import PaymentServiceModel
from model.repository import PosRepositoryModel

VOUCHER = 3
DEBIT = 2


class PaymentService(PaymentServiceModel, BaseRepository):

    def __init__(
        self,
        message_bus,        # type: MbContextMessageBus
        config,             # type: Config
        pos_repository,     # type: PosRepositoryModel
    ):
        # type: (...) -> None

        super(PaymentService, self).__init__(message_bus.mbcontext)
        self.message_bus = message_bus
        self.config = config
        self.payment_types = None
        self.pos_repository = pos_repository

    def get_payment_types(
        self
    ):
        # type: () -> List[PaymentType]

        def inner_func(conn):
            # type: (Connection) -> List[int]

            query = """
                SELECT
                    TT.TenderId AS TenderId,
                    TT.TenderDescr AS TenderDescr,
                    TT.ElectronicType AS ElectronicTypeId,
                    TET.TypeDescr AS ElectronicTypeDescr,
                    TT.ParentTenderId AS ParentTenderId,
                    TT.CustomAttr As CustomAttr
                FROM productdb.TenderType TT
                JOIN productdb.TenderElectronicType TET ON TET.TypeId=TT.ElectronicType;
            """
            cursor = conn.select(query)

            tenders = {}
            for row in cursor:
                tender_id = int(row.get_entry('TenderId'))
                parent_tender_id = row.get_entry('ParentTenderId')
                tenders[tender_id] = {
                    'id': tender_id,
                    'descr': row.get_entry('TenderDescr'),
                    'electronicType': row.get_entry('ElectronicTypeDescr'),
                    'electronicTypeId': int(row.get_entry('ElectronicTypeId')),
                    'parentId ': int(parent_tender_id) if parent_tender_id else None,
                    'customAttr': row.get_entry('CustomAttr')
                }

            return tenders

        if not self.payment_types:
            self.payment_types = self.format_payment_types(self.execute_with_connection(
                inner_func,
                service_name="Persistence"
            ))

        return self.payment_types

    def format_payment_types(
        self,
        tender_types,    # type: List[Dict[int, str]]
    ):
        # type: (...) -> List[PaymentType]

        payment_types = []
        config_file = open(self.config.totem_config_path, 'rb')
        data = json.load(config_file)
        payments = data.get("payments", {})
        tenders = [tender for _, tender in tender_types.items()]

        for key, payment in payments.items():
            electronic_type_json_id = int(key)
            tender = [tender for tender in tenders if electronic_type_json_id == tender.get("electronicTypeId", -1)]
            payment_type_id = None

            if len(tender) == 0:
                if electronic_type_json_id != VOUCHER:
                    continue

                payment_type_id = DEBIT
                electronic_type_json_id = DEBIT

            if payment_type_id is None:
                payment_type_id = tender[0].get("id")

            payment_name = payment.get("paymentName", {})
            payment_name = self._get_payment_names(payment_name=payment_name)
            image_path = payment.get("name", "")

            payment_type = PaymentType(
                payment_type=payment_type_id,
                payment_name=payment_name,
                image_path=image_path,
                electronic_type_id=electronic_type_json_id,
            )

            payment_types.append(payment_type)

        if not payment_types:
            raise ValidationError(PosValidationErrorsType.PaymentTypesToShowNotFound)

        return sorted(payment_types, key=lambda p: p.payment_type)

    def get_tender_details(
        self,
        payment_data,   # type: OrderInsertPaymentRequest
    ):
        # type: (...) -> str

        tender_details = json.dumps({
            "Media": self._validate_required_property(
                property_name="media",
                property_value=payment_data.media,
                check_required_property=True,
            ),
            "CNPJAuth": self._validate_required_property(
                property_name="cnpj_auth",
                property_value=payment_data.cnpj_auth,
                check_required_property=True,
            ),
            "TransactionProcessor": self._validate_required_property(
                property_name="transaction_processor",
                property_value=payment_data.transaction_processor,
                check_required_property=True,
            ),
            "Bandeira": self._validate_required_property(
                property_name="brand",
                property_value=payment_data.brand,
                check_required_property=True,
            ),
            "IdAuth": self._validate_required_property(
                property_name="id_auth",
                property_value=payment_data.id_auth,
                check_required_property=True,
            ),
            "AuthCode": self._validate_required_property(
                property_name="auth_code",
                property_value=payment_data.auth_code,
                check_required_property=True,
            ),
            "ReceiptCustomer": self._validate_required_property(
                property_name="receipt_customer",
                property_value=payment_data.receipt_customer,
                check_required_property=True,
            ),
            "ReceiptMerchant": self._validate_required_property(
                property_name="receipt_merchant",
                property_value=payment_data.receipt_merchant,
                check_required_property=True,
            ),
            "AccreditorSat": self._validate_required_property(
                property_name="accreditor_sat",
                property_value=payment_data.accreditor_sat,
                check_required_property=False,
            ),
            "tBand": self._validate_required_property(
                property_name="t_band",
                property_value=payment_data.t_band,
                check_required_property=False,
            ),
            "tPag": self._validate_required_property(
                property_name="t_pag",
                property_value=payment_data.t_pag,
                check_required_property=False,
            ),
            "NsuHost": self._validate_required_property(
                property_name="nsu_host",
                property_value=payment_data.nsu_host,
                check_required_property=False,
            ),
            "CardNumber": self._validate_required_property(
                property_name="card_number",
                property_value=payment_data.card_number,
                check_required_property=False,
            ),
            "TransactionIdentifier":  self._validate_required_property(
                property_name="transaction_identifier",
                property_value=payment_data.transaction_identifier,
                check_required_property=True,
            )
        })

        return tender_details

    @staticmethod
    def _validate_required_property(
        property_name,                      # type: str
        property_value,                     # type: str
        check_required_property=False,      # type: bool
    ):
        # type: (...)-> str

        if check_required_property and property_value in [None, ""]:
            raise ValidationError(
                PosValidationErrorsType.PropertyRequiredNotFound,
                property_name,
            )

        return property_value

    def _get_payment_names(
        self,
        payment_name,   # type: Dict
    ):
        # type: (...) -> Dict

        languages = self.pos_repository.get_totem_languages()
        payment_name_dict = {}

        for language in languages:
            payment_name_dict.update(
                {language: payment_name.get(language) or ''}
            )

        return payment_name_dict



from typing import (
    Dict,
    Optional,
)


class OrderDiscountsDetail(object):
    def __init__(
        self,
        discount_ids=None,         # type: Optional[Dict]
        discount_amounts=None,     # type: Optional[Dict]
    ):
        # type: (...) -> None

        self.discount_ids = discount_ids if discount_ids else dict()
        self.discount_amounts = discount_amounts if discount_amounts else dict()



import json
from xml.etree import cElementTree as eTree

from typing import (
    Optional,
    Tuple,
)

from application.service import OrderService

from application.interactor import OnOrderReopenInteractor
from application.model import Configurations
from messagebus import DefaultToken, DataType, Message
from messageprocessorutil import UuidMessageProcessor


class OnOrderReopenProcessor(UuidMessageProcessor):
    def __init__(self, configs, interactor, order_service):
        # type: (Configurations, OnOrderReopenInteractor, OrderService) -> None

        super(OnOrderReopenProcessor, self).__init__()
        self.logger = configs.logger
        self.configs = configs
        self.interactor = interactor
        self.order_service = order_service

    def parse_data(
        self,
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (...) -> (int, Optional[str])

        data = json.loads(data)
        pos_id = int(data.get("pos_id"))
        table_id = data.get("table_id") or None

        return pos_id, table_id

    def call_business(
        self,
        params,     # type: Tuple[int, str]
    ):
        # type: (...) -> Optional[eTree.Element]

        pos_id, table_id = params
        return self.interactor.execute(
            pos_id=pos_id,
            table_id=table_id,
        )

    def format_response(self, message_bus, message, event, request, response):
        if message:
            data = eTree.tostring(element=response) if response is not None else ""
            message_bus.reply_message(
                message=message,
                reply_message=Message(
                    token=DefaultToken.TK_SYS_ACK,
                    data_type=DataType.string,
                    data=data,
                ),
            )

    def format_exception(self, message_bus, message, event, input_obj, exception):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



# -*- coding: utf-8 -*-
import json
from datetime import datetime, date
from typing import List, Union, Dict
from xml.etree import cElementTree as eTree

from application.configurator import Configurator
from application.model import EventData
from iso8601 import parse_date


class ProductionOrderInteractor(object):

    def __init__(self, configurations):
        # type: (Configurator) -> None

        self.configs = configurations
        self.production_order = None

    def build(self, event_data):
        # type: (EventData) -> str

        self.production_order = event_data.data.find("ProductionOrder") or event_data.data

        payload = {
            "created-at": self.production_order.get("created_at_gmt") + "Z",
            "operator-id": int(self.production_order.get("operator_id")),
            "operator-level": int(self.production_order.get("operator_level")),
            "operator-name": self.production_order.get("operator_name"),
            "order-id": int(self.production_order.get("order_id")),
            "originator": self.production_order.get("originator"),
            "pod-type": self.production_order.get("pod_type"),
            "pos-id": int(self.production_order.get("pos_id")),
            "prod-state": self.production_order.get("prod_state"),
            "sale-type": self.production_order.get("sale_type"),
            "session-id": self.production_order.get("session_id"),
            "state": self.production_order.get("state"),
            "state-id": int(self.production_order.get("state_id")),
            "store-code": str(self.configs.store_id),
            "store-division": str(self.configs.store_division) if self.configs.store_division else None,
            "total-gross": float(self.production_order.get("total_gross")),
            "tagged-lines": self._get_tagged_lines(),
            "items": self._get_items(),
            "properties": self._get_properties(),
            "state-history": self._get_state_history(),
            "prod-state-history": self._get_prod_state_history()
        }

        return json.dumps(payload)

    def _get_tagged_lines(self):
        # type: () -> Dict

        parsed_tagged_lines = self.production_order.get("tagged_lines").split(",")
        tagged_lines = {}

        for tagged_line in parsed_tagged_lines:
            items_tags = tagged_line.split("=")
            if len(items_tags) < 2:
                self.configs.logger.warning("tagged_line with less than two parameters: {}".format(tagged_line))
                continue

            item, tag = items_tags
            if item not in tagged_lines:
                tagged_lines[item] = [tag]
            else:
                tagged_lines[item].append(tag)

        return tagged_lines

    def _get_items(self):
        # type: () -> List

        properties = []

        for item in self.production_order.find("Items"):
            data = self._build_item_payload(item)
            properties.append(data)

        return properties

    def _build_item_payload(self, item):
        # type: (eTree) -> Dict

        data = {
            "context-id": item.get("item_code"),
            "default-qty": float(item.get("default_qty")),
            "description": item.get("description"),
            "item-type": item.get("item_type"),
            "level": int(item.get("level")),
            "line-number": int(item.get("line_number")),
            "max-qty": float(item.get("max_qty")),
            "min-qty": float(item.get("min_qty")),
            "multiplied-qty": float(item.get("multiplied_qty")),
            "part-code": int(item.get("part_code")),
            "product-priority": int(item.get("product_priority")),
            "qty": float(item.get("qty")),
            "qty-added": float(item.get("qty_added")),
            "qty-modified": float(item.get("qty_modified")),
            "qty-voided": float(item.get("qty_voided")),
            "product-tags": self._build_item_json_tags(item),
            "tags": self._build_item_tags(item),
            "tag-history": self._build_item_tag_history(item),
            "jit-lines": item.get("jit_lines"),
            "items": [self._build_item_payload(sub_item) for sub_item in item.findall("Item")]
        }

        return data

    @staticmethod
    def _build_item_json_tags(item):
        # type: (eTree) -> Dict

        json_tags = json.loads(item.get("json_tags") or "{}")
        tags = {}

        for tag in json_tags:
            key, value = tag.split("=")
            tags[str(key)] = str(value)

        return tags

    def _build_item_tags(self, item):
        # type: (eTree) -> List

        tags = []

        for tag in self._find(item, "Tags"):
            tags.append(str(tag.get("name")))

        return tags

    def _build_item_tag_history(self, item):
        # type: (eTree) -> List

        tags = []

        for tag_event in self._find(item, "TagHistory"):
            data = {
                "tag": tag_event.get("tag"),
                "action": tag_event.get("action"),
                "timestamp": self._parse_timestamp(tag_event.get("timestamp"))
            }
            tags.append(data)

        return tags

    def _get_properties(self):
        # type: () -> Dict

        properties = {}

        for prop in self._find(self.production_order, "Properties"):
            properties[prop.get("key")] = prop.get("value")

        return properties

    def _get_state_history(self):
        # type: () -> List

        state_history = []

        for prop in self._find(self.production_order, "StateHistory"):
            data = {
                "state": prop.get("state"),
                "state-id": int(prop.get("state_id")),
                "timestamp": self._parse_timestamp(prop.get("timestamp"))
            }
            state_history.append(data)

        return state_history

    def _get_prod_state_history(self):
        # type: () -> List

        prod_state_history = []

        for prop in self._find(self.production_order, "ProdStateHistory"):
            prod_state = prop.get("prod_state")
            if prod_state is None:
                continue

            data = {
                "prod-state": prod_state,
                "timestamp": self._parse_timestamp(prop.get("timestamp"))
            }
            prod_state_history.append(data)

        return prod_state_history

    @staticmethod
    def _find(element, prop):
        # type: (eTree, str) -> Union[eTree, List]

        return element.find(prop) or []

    def _parse_timestamp(self, timestamp):
        # type: (str) -> str

        parsed_timestamp = parse_date(timestamp).replace(tzinfo=None, microsecond=0)
        converted_time = self._convert_to_utc(parsed_timestamp)

        return converted_time.isoformat()

    @staticmethod
    def _convert_to_utc(local_datetime):
        # type: (date) -> date

        time_diff = datetime.utcnow() - datetime.now()
        converted_time = local_datetime + time_diff

        return converted_time



import datetime

from commons.dto import ReportColumnDefinition, AlignTypes
from commons.report import Formatter
from commons.util import TableReport
from reports_app.closedaysummary.dto import CloseDaySummaryDto  # noqa
from reports_app.closedaysummary.dto import CloseDaySummaryHeaderDto  # noqa


class CloseDayFormatter(Formatter):
    def __init__(self):
        self.report_lines = []
        self.translated_messages = []

    def format_report(self, close_day_summary_dto, translator):
        # type: (CloseDaySummaryDto) -> None
        self.translated_messages = translator.translate_labels(
            labels=[
                "$CLOSE_DAY",
                "$STORE",
                "$LABEL_POS",
                "$OPERATOR",
                "$CURRENT_DATE",
                "$BUSINESS_DATE",
                "$OPEN_DAY",
                "$AUTHORIZER",
                "$TYPE",
                "$QTY",
                "$VALUE",
                "$INITIAL_BALANCE",
                "$SALES",
                "$VOIDED_ORDERS",
                "$DISCOUNTS",
                "$DONATIONS",
                "$LIQUID_SALES",
                "$TENDER_TYPES",
                "$CASH_OUT",
                "$CASH_IN",
                "$VALUE_IN_DRAWER",
            ],
            pos_id=close_day_summary_dto.close_day_summary_header_dto.pos_id
        )

        report = self._append_header(close_day_summary_dto.close_day_summary_header_dto,
                                     close_day_summary_dto.close_day_summary_body_dto)
        report += self._append_body(close_day_summary_dto.close_day_summary_body_dto)

        report_bytes = report.decode("utf-8").encode("utf-8")
        return report_bytes

    def _append_header(self, close_day_summary_header_dto, close_day_summary_body_dto):
        # type: (CloseDaySummaryHeaderDto, CloseDaySummaryBodyDto) -> str
        self._append_line_separator()
        self._append_title()
        self._append_store_pos(close_day_summary_header_dto.store_id, close_day_summary_header_dto.pos_id)
        self._append_operator(close_day_summary_header_dto.operator_id)
        self._append_current_date()
        self._append_business_date(close_day_summary_header_dto.business_date)
        self._append_empty_line()

        self._append_open_day_time(close_day_summary_body_dto.open_day_time)
        self._append_authorizer(close_day_summary_body_dto.authorizer)

        table_formatter = TableReport([
            ReportColumnDefinition(width=17, fill_with_char='.'),
            ReportColumnDefinition(width=21, before_text=': ')
        ])

        return table_formatter.format(self.report_lines)

    def _append_body(self, close_day_summary_body_dto):
        self.report_lines = []

        self._append_line_separator()

        self._append_report_body_title(self.report_lines)
        self._append_body_items(close_day_summary_body_dto, self.report_lines)
        self._append_line_separator()
        return self._format_body(self.report_lines)

    def _append_line_separator(self):
        self.report_lines.append([ReportColumnDefinition(
            fill_with_char='='
        )])

    def _append_title(self):
        title = self.translated_messages["$CLOSE_DAY"]

        self.report_lines.append([ReportColumnDefinition(
            text=title,
            align=AlignTypes.CENTER
        )])

    def _append_store_pos(self, store_id, pos_id):
        self.report_lines.append([self.translated_messages["$STORE"] + "/" + self.translated_messages["$LABEL_POS"],
                                  (str(store_id) + '/' + str(pos_id))])

    def _append_operator(self, operator_id):
        self.report_lines.append([self.translated_messages["$OPERATOR"], str(operator_id)])

    def _append_current_date(self):
        self.report_lines.append([self.translated_messages["$CURRENT_DATE"],
                                  datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")])

    def _append_open_day_time(self, open_day_time):
        self.report_lines.append([self.translated_messages["$OPEN_DAY"], str(open_day_time)])

    def _append_business_date(self, business_date):
        self.report_lines.append([self.translated_messages["$BUSINESS_DATE"], business_date])

    def _append_empty_line(self):
        self.report_lines.append([ReportColumnDefinition(fill_with_char=' ')])

    def _append_authorizer(self, authorizer):
        self.report_lines.append([self.translated_messages["$AUTHORIZER"], str(authorizer)])

    def _append_body_items(self, logout_summary_body_dto, report_body_lines):
        for line in logout_summary_body_dto.report_body:
            self._append_transfers_info(line, report_body_lines)

    def _append_transfers_info(self, report_item, report_body_lines):
        line_type = self.translated_messages["" if report_item.name is None else report_item.name]
        separator_1 = "" if report_item.quantity is None else ": ["

        if report_item.quantity is None:
            quantity = ""
            separator_1 = ":  "
        else:
            quantity = str(report_item.quantity)

        if report_item.value is None and report_item.quantity is None:
            separator_2 = ""
        elif report_item.quantity is None:
            separator_2 = "  R$"
        elif report_item.value is None:
            separator_2 = "]   "
        else:
            separator_2 = "] R$"

        if report_item.value is None:
            value = ""
        else:
            value = str(report_item.value)

        report_body_lines.append([line_type,
                                  separator_1,
                                  quantity,
                                  separator_2,
                                  value])

    def _append_report_body_title(self, report_body_lines):
        report_body_lines.append([
            ReportColumnDefinition(text=self.translated_messages["$TYPE"], width=17),
            ReportColumnDefinition(width=3),
            ReportColumnDefinition(text=self.translated_messages["$QTY"], width=4, align=AlignTypes.RIGHT),
            ReportColumnDefinition(width=4),
            ReportColumnDefinition(text=self.translated_messages["$VALUE"] + "(R$)", width=10, align=AlignTypes.RIGHT)
        ])

    @staticmethod
    def _format_body(report_body_lines):
        table_formatter = TableReport([
            ReportColumnDefinition(width=17, fill_with_char='.'),
            ReportColumnDefinition(text=': [', width=3),
            ReportColumnDefinition(width=4, align=AlignTypes.RIGHT),
            ReportColumnDefinition(text='] R$', width=4),
            ReportColumnDefinition(text='', width=10, align=AlignTypes.RIGHT)
        ])
        return table_formatter.format(report_body_lines)



from typing import (
    Optional,
    Any,
)

from messagebus import (
    Message,
    DefaultToken,
    DataType,
    MessageBus,
    TokenPriority,
    TokenCreator,
    Event,
)

from messageprocessorutil import UuidMessageProcessor

from messagebus import AckMessage, NakMessage


class PrnStatusPrintProcessor(UuidMessageProcessor):
    def __init__(
        self,
        pos_id,         # type: int
        message_bus,    # type: MessageBus
    ):
        # type: (...) -> None
        super(PrnStatusPrintProcessor, self).__init__()

        self.pos_id = pos_id
        self.message_bus = message_bus

    def parse_data(
        self,
        data,           # type: Optional[str]
        event_type,     # type: Optional[str]
    ):
        # type: (...) -> Optional[str]

        return data

    def call_business(
        self,
        receive_data,      # type: Optional[str]
    ):
        return "MOBILE_PRINT"

    def format_response(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        request,        # type: Optional[str]
        result,         # type: str
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=AckMessage(data=result),
        )

    def format_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        input_obj,      # type: Optional[str]
        exception,      # type: Exception
    ):
        # type: (...) -> None

        self.format_parse_exception(
            message_bus=message_bus,
            message=message,
            event=event,
            data=input_obj,
            exception=exception,
        )

    @staticmethod
    def format_parse_exception(
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        data,           # type: Optional[str]
        exception,      # type: Exception
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=NakMessage(data=exception.message),
        )



from _TefCancellationReportGenerator import TefCancellationReportGenerator
from _TefCancellationReportPresenter import TefCancellationReportPresenter

__all__ = ["TefCancellationReportGenerator", "TefCancellationReportPresenter"]



from typing import (
    Dict,
    List,
    Optional,
)


class ActionResult(object):
    def __init__(
        self,
        discounts_id,                   # type: List
        items_to_burn,                  # type: Dict
        order_tender_ids,               # type: List
        discount_item_values,           # type: Dict
        benefit_applied,                # type: bool
        pending_to_apply,               # type: bool
        benefit_ids_to_remove=None,     # type: Optional[List[str]]
    ):
        # type: (...) -> None

        self.discounts_id = discounts_id
        self.items_to_burn = items_to_burn
        self.order_tenders = order_tender_ids
        self.discount_item_values = discount_item_values
        self.benefit_applied = benefit_applied
        self.pending_to_apply = pending_to_apply
        self.benefit_ids_to_remove = benefit_ids_to_remove if benefit_ids_to_remove else []

    def __add__(self, action_result):
        # type: (ActionResult) -> ActionResult

        return self._merge_action_result(self, action_result)

    def __radd__(self, action_result):
        # type: (ActionResult) -> ActionResult

        return self._merge_action_result(self, action_result)

    def __nonzero__(self):
        # type: () -> bool
        return bool(self.discounts_id) or bool(self.items_to_burn) or bool(self.order_tenders) or self.pending_to_apply

    @staticmethod
    def _merge_action_result(current, new):
        current.discounts_id.extend(new.discounts_id)
        current.order_tenders.extend(new.order_tenders)
        current.discount_item_values.update(new.discount_item_values)
        current.benefit_applied = current.benefit_applied and new.benefit_applied
        current.pending_to_apply = current.pending_to_apply or new.pending_to_apply
        current.benefit_ids_to_remove.extend(new.benefit_ids_to_remove)

        return ActionResult(
            discounts_id=current.discounts_id,
            items_to_burn=current.items_to_burn,
            order_tender_ids=current.order_tenders,
            discount_item_values=current.discount_item_values,
            benefit_applied=current.benefit_applied,
            pending_to_apply=current.pending_to_apply,
            benefit_ids_to_remove=current.benefit_ids_to_remove,
        )




from sysactions import (
    show_any_dialog, action,
)

@action
def do_start_network_test(
    pos_id,  # type: str
):
    # type: (...) -> None
    show_any_dialog(
        pos_id,
        "networkTest",
        "",
        "",
        "",
        "",
        "",
        60000,
        "",
        "",
        "",
        "",
        "",
        True
    )



from utils import center, double_line_separator
from utils.report_formatters import translate_pos_label

DATE_TIME_FMT = "%d/%m/%Y %H:%M:%S"


def report_default_header(title, pos_list, operator_id, period, current_datetime, store_id=None):
    title = center(title)
    separator = double_line_separator()
    pos_list = [pos_list] if type(pos_list) != list else pos_list
    selected_pos = "Todos" if type(pos_list) == list and len(pos_list) > 1 else pos_list[0]
    pos = translate_pos_label(pos_id=selected_pos)

    header = separator
    header += title
    if store_id:
        header += "Loja..........: {}".format(store_id.upper()) + "\n"
    header += "Data/hora.....: {}".format(current_datetime) + "\n"

    if period is not None:
        fiscal_date = "%02d/%02d/%04d" % (int(period[6:8]), int(period[4:6]), int(period[:4]))
        header += "Data fiscal...: {}".format(fiscal_date) + "\n"

    if operator_id is not None and selected_pos is not None:
        operator = operator_id if len(operator_id) > 0 else "Todos"
        header += "ID Operador ..: {} ({} # {})".format(operator, pos, selected_pos) + "\n"

    header += separator

    return header



from _OrderRepository import OrderRepository
from _AccountRepository import AccountRepository
from _TenderRepository import TenderRepository
from _FiscalRepository import FiscalRepository
from _PosCtrlRepository import PosCtrlRepository
from _TableRepository import TableRepository
from _ProductRepository import ProductRepository

__all__ = ["OrderRepository",
           "AccountRepository",
           "TenderRepository",
           "FiscalRepository",
           "PosCtrlRepository",
           "TableRepository",
           "ProductRepository"]



from datetime import datetime
from typing import Optional

from application.model import (
    BenefitAppliers,
    OperationDescription,
)


class BenefitControllerDto(object):

    def __init__(
        self,
        benefit_id=None,                # type: Optional[str]
        operation_description=None,     # type: Optional[OperationDescription]
        benefit_applier=None,           # type: Optional[BenefitAppliers]
        retry_quantity=None,            # type: Optional[int]
        last_retry_date_utc=None,       # type: Optional[datetime]
        order_picture=None,             # type: Optional[datetime]
        table_picture=None,             # type: Optional[datetime]
    ):
        # type: (...) -> None

        self.benefit_id = benefit_id
        self.operation_description = operation_description
        self.benefit_applier = benefit_applier
        self.retry_quantity = retry_quantity
        self.last_retry_date_utc = last_retry_date_utc
        self.order_picture = order_picture
        self.table_picture = table_picture




import logging
from datetime import datetime

from messagebus import (
    MessageBus,
)

from devicemanager.model import DeviceStatusHistory
from devicemanager.repository import DeviceManagerRepository
from devicemanager.service import (
    GetDevicesService,
    DeviceStatusService,
)
from .dto import (
    UpdateDevicesStatusRequest,
    UpdateDevicesStatusResponse,
)

logger = logging.getLogger(__name__)


class UpdateDevicesStatusInteractor(object):
    def __init__(
        self,
        repository,             # type: DeviceManagerRepository
        get_devices_service,    # type: GetDevicesService
        device_status_service,  # type: DeviceStatusService
        message_bus,            # type: MessageBus
    ):
        # type: (...) -> None

        self.repository = repository
        self.get_devices_service = get_devices_service
        self.device_status_service = device_status_service
        self.message_bus = message_bus

    def update_devices(
        self,     # type: UpdateDevicesStatusInteractor
        request,  # type: UpdateDevicesStatusRequest
    ):
        # type: (...) -> UpdateDevicesStatusResponse

        change_date = datetime.utcnow()
        device_names_to_remove = []
        device_names_to_add = []
        disabled_devices = self.repository.get_disabled_devices()
        for device in request.devices:
            if device.enabled and device.name in disabled_devices:
                device_names_to_remove.append(device.name)
            elif not device.enabled and device.name not in disabled_devices:
                device_names_to_add.append(device.name)

        updated = False
        for device_name in device_names_to_add:
            logger.info("Turning off {}".format(device_name))
            self.repository.create_device_status(device_name=device_name)
            self._create_status_history(
                device_name=device_name,
                change_date=change_date,
                operator=request.operator,
                enabled=False,
            )
            updated = True
        for device_name in device_names_to_remove:
            logger.info("Turning on {}".format(device_name))
            self.repository.delete_device_status(device_name=device_name)
            self._create_status_history(
                device_name=device_name,
                change_date=change_date,
                operator=request.operator,
                enabled=True,
            )
            updated = True

        devices = self.get_devices_service.get_devices()
        self.device_status_service.initialize_cache(repository=self.repository)
        return UpdateDevicesStatusResponse(
            updated=updated,
            devices=devices,
        )

    def _create_status_history(
        self,           # type: UpdateDevicesStatusInteractor
        device_name,    # type: str
        change_date,    # type: datetime
        operator,       # type: str
        enabled,        # type: bool
    ):
        self.repository.create_device_status_history(
            device_name=device_name,
            device_status_history=DeviceStatusHistory(
                enabled=enabled,
                timestamp=change_date,
                operator=operator,
            ),
        )



from logging import Logger
from memorycache import (
    CacheManager,
    PeriodBasedExpiration,
)
from typing import (
    Any,
    Dict,
    Optional,
)

from httpmessageprocessor import HttpMessageProcessorMessageHandler
from interactor.event import ElectronicPaymentEventInteractor
from interactor.route.electronic_payment.cache import CacheInteractor
from interactor.route.electronic_payment.cache.dto import CacheRequest
from interactor.route.customization.get_styles import GetStylesInteractor
from interactor.route.customization.get_styles.dto import GetStylesRequest
from interactor.route.loyalty.search_customer_info import SearchCustomerInfoInteractor
from interactor.route.loyalty.search_customer_info.dto import SearchCustomerInfoRequest
from interactor.route.menu.get_images import GetImagesInteractor
from interactor.route.menu.get_images.dto import GetImagesRequest
from interactor.route.menu.get_navigation import GetNavigationInteractor
from interactor.route.menu.get_navigation.dto import GetNavigationRequest
from interactor.route.menu.get_product_info import GetProductInfoInteractor
from interactor.route.menu.get_product_info.dto import GetProductInfoRequest
from interactor.route.menu.get_products import GetProductsInteractor
from interactor.route.menu.get_products.dto import GetProductsRequest
from interactor.route.menu.get_ruptured_products import GetRupturedProducts
from interactor.route.menu.get_ruptured_products.dto import GetRupturedProductsRequest
from interactor.route.menu.get_banners import GetBannersInteractor
from interactor.route.menu.get_banners.dto import GetBannersRequest
from interactor.route.menu.get_up_sell_part_codes import GetUpSellPartCodesInteractor
from interactor.route.menu.get_up_sell_part_codes.dto import GetUpSellPartCodesRequest
from interactor.route.order.add_items import AddItemsInteractor
from interactor.route.order.add_items.dto import AddItemsRequest
from interactor.route.order.add_properties import AddPropertiesInteractor
from interactor.route.order.add_properties.dto import AddPropertiesRequest
from interactor.route.order.apply_benefit import ApplyBenefitInteractor
from interactor.route.order.apply_benefit.dto import ApplyBenefitRequest
from interactor.route.order.apply_coupon import ApplyCouponInteractor
from interactor.route.order.apply_coupon.dto import ApplyCouponRequest
from interactor.route.order.cancel import OrderCancelInteractor
from interactor.route.order.cancel.dto import OrderCancelRequest
from interactor.route.order.change_items_quantity import ChangeItemsQuantityInteractor
from interactor.route.order.change_items_quantity.dto import ChangeItemsQuantityRequest
from interactor.route.order.close_shopping_cart import CloseShoppingCartInteractor
from interactor.route.order.close_shopping_cart.dto import CloseShoppingCartRequest
from interactor.route.order.create import OrderCreateInteractor
from interactor.route.order.create.dto import OrderCreateRequest
from interactor.route.order.delete_line import OrderDeleteLineInteractor
from interactor.route.order.delete_line.dto import OrderDeleteLineRequest
from interactor.route.order.finalize import OrderFinalizeInteractor
from interactor.route.order.finalize.dto import OrderFinalizeRequest
from interactor.route.order.get_coupons import GetCouponInteractor
from interactor.route.order.get_coupons.dto import GetCouponRequest
from interactor.route.order.get_detail import OrderGetDetailInteractor
from interactor.route.order.get_detail.dto import OrderGetDetailRequest
from interactor.route.order.save import OrderSaveInteractor
from interactor.route.order.save.dto import OrderSaveRequest
from interactor.route.payment.insert import OrderInsertPaymentInteractor
from interactor.route.payment.insert.dto import OrderInsertPaymentRequest
from interactor.route.order.reopen_shopping_cart import ReopenShoppingCartInteractor
from interactor.route.order.reopen_shopping_cart.dto import ReopenShoppingCartRequest
from interactor.route.order.remove_coupon import RemoveCouponInteractor
from interactor.route.order.remove_coupon.dto import RemoveCouponRequest
from interactor.route.payment.get_payment_types import GetPaymentTypesInteractor
from interactor.route.payment.get_payment_types.dto import GetPaymentTypesRequest
from interactor.route.electronic_payment.read_messages import ReadMessagesInteractor
from interactor.route.electronic_payment.read_messages.dto import ReadMessagesRequest
from interactor.route.electronic_payment.clear_messages import ClearMessagesInteractor
from interactor.route.electronic_payment.clear_messages.dto import ClearMessagesRequest
from interactor.route.electronic_payment.send_responses import SendResponsesInteractor
from interactor.route.electronic_payment.send_responses.dto import SendResponsesRequest
from interactor.route.customization.get_custom_images import GetCustomImagesInteractor
from interactor.route.customization.get_custom_images.dto import GetCustomImagesRequest
from interactor.route.customization.get_text import GetTextInteractor
from interactor.route.customization.get_text.dto import GetTextRequest
from interactor.route.peripheral.print_text import PrintTextInteractor
from interactor.route.peripheral.print_text.dto import PrintTextRequest
from interactor.route.peripheral.read_scanner import ReadScannerInteractor
from interactor.route.peripheral.read_scanner.dto import ReadScannerRequest
from interactor.route.store.get_status import GetStoreInteractor
from interactor.route.store.get_status.dto import GetStoreRequest
from interactor.route.store.get_config import GetStoreConfigInteractor
from interactor.route.store.get_config.dto import GetStoreConfigRequest
from interactor.route.payment.cancel import CancelPaymentInteractor
from interactor.route.payment.cancel.dto import CancelPaymentRequest
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, Event
from messagehandler import MessageHandlerBuilder, MessageHandler
from messageprocessor import (
    MessageProcessor,
    DefaultMessageProcessorExecutorFactory,
    MessageProcessorMessageHandler,
)
from messageprocessorutil import LoggingProcessorCallback
from model.config import Config
from model.event import ListenedEvents
from model.exception import (
    AuthValidationErrors,
    BadRequestError,
    ClientCustomError,
    ExternalServiceErrors,
    InUseError,
    NotAvailableError,
    NotFoundError,
    OperationErrors,
    ValidationError,
    BlockedError,
)
from model.response import Responses
from processor.event.electronic_payment import (
    ElectronicPaymentCloseDialogProcessor,
    ElectronicPaymentShowQRCodeProcessor,
    ElectronicPaymentShowChoicesProcessor,
    ElectronicPaymentShowInputProcessor,
    ElectronicPaymentShowMessageProcessor,
    ElectronicPaymentShowQuestionProcessor
)
from processor.event.pos_logger import PosLoggerAutoMappedMessageProcessor
from pyscripts_cache import PyscriptsCache
from repository import PosRepository
from services import (
    FiscalService,
    OrderService,
    ApiService,
    CouponService,
    ImagesService,
    PaymentService,
    ProductInfoService,
    PrinterService,
    LoyaltyService,
    NavigationService,
)
from services._ElectronicPaymentService import ElectronicPaymentService

from simplehttprouter import Route


class ServiceMessageHandlerBuilder(MessageHandlerBuilder):

    def __init__(self, logger, config, message_bus, flask_server=False):
        # type: (Logger, Config, MbContextMessageBus, bool) -> None

        self.navigation_service = None
        self.logger = logger
        self.config = config
        self.message_bus = message_bus
        self.flask_server = flask_server

        self.pos_repository = None
        self.fiscal_service = None
        self.coupon_service = None
        self.order_service = None
        self.api_service = None
        self.modifier_service = None
        self.images_service = None
        self.payment_service = None
        self.product_info_service = None
        self.printer_service = None
        self.loyalty_service = None
        self.electronic_payment_service = None

        self.pyscripts_cache = None

        self.cache_menu_ruptured_items = {"last_request_date": None, "data": None}
        self.cache_menu_products = None
        self.memory_cache = self._build_memory_cache()
        self.cache_menu_navigation = []
        self.electronic_messages = []

        self.callbacks = None
        self.message_handler = None
        self.logger_created = {}

    def build_singletons(self):
        # type: () -> None

        self.callbacks = [LoggingProcessorCallback(self.logger)]

        if self.flask_server:
            self.navigation_service = NavigationService(self.message_bus)
            self.printer_service = PrinterService(
                logger=self.logger,
                message_bus=self.message_bus,
            )
            self.pos_repository = PosRepository(
                self.message_bus,
                self.logger,
                self.config,
                self.navigation_service,
            )
            self.fiscal_service = FiscalService(
                message_bus=self.message_bus,
                logger=self.logger,
                pos_repository=self.pos_repository,
                printer_service=self.printer_service,
            )
            self.coupon_service = CouponService(self.message_bus, self.logger)
            self.pyscripts_cache = PyscriptsCache(self.message_bus.mbcontext)
            self.loyalty_service = LoyaltyService(
                message_bus=self.message_bus,
                config=self.config,
                logger=self.logger,
            )
            self.order_service = OrderService()
            self.images_service = ImagesService()
            self.payment_service = PaymentService(
                message_bus=self.message_bus,
                config=self.config,
                pos_repository=self.pos_repository,
            )
            self.product_info_service = ProductInfoService(self.message_bus)
            self.fiscal_service.get_electronic_payments_id()

            self.electronic_payment_service = ElectronicPaymentService(
                logger=self.logger,
                message_bus=self.message_bus,
                electronic_payments_id=self.fiscal_service.electronic_payments_id,
            )
        else:
            self.api_service = ApiService(self.config, self.logger)

            events_processors = self._build_events_processors()
            default_message_processor_executor_factory = DefaultMessageProcessorExecutorFactory(self.callbacks)
            self.message_handler = MessageProcessorMessageHandler(events_processors,
                                                                  None,
                                                                  default_message_processor_executor_factory)

    def build_message_handler(self, message=None, event=None):
        # type: (Optional[Message], Optional[Event]) -> MessageHandler

        if self.flask_server:
            processors_routes = self._build_processors_routes()
            http_message_processor = HttpMessageProcessorMessageHandler(route_processor=processors_routes,
                                                                        callbacks=self.callbacks)
            return http_message_processor
        else:
            return self.message_handler

    def destroy_message_handler(self, message_handler):
        # type: (MessageHandler) -> None

        pass

    def destroy_singletons(self):
        # type: () -> None

        pass

    def _build_events_processors(self):
        # type: () -> Optional[Dict[str, MessageProcessor]]

        electronic_payment_event_interactor = ElectronicPaymentEventInteractor(self.api_service)

        electronic_payment_close_dialog_processor = \
            ElectronicPaymentCloseDialogProcessor(electronic_payment_event_interactor)
        electronic_payment_show_qrcode_processor = \
            ElectronicPaymentShowQRCodeProcessor(electronic_payment_event_interactor)
        electronic_payment_show_choices_processor = \
            ElectronicPaymentShowChoicesProcessor(electronic_payment_event_interactor)
        electronic_payment_show_input_processor = \
            ElectronicPaymentShowInputProcessor(electronic_payment_event_interactor)
        electronic_payment_show_message_processor = \
            ElectronicPaymentShowMessageProcessor(electronic_payment_event_interactor)
        electronic_payment_show_question_processor = \
            ElectronicPaymentShowQuestionProcessor(electronic_payment_event_interactor)

        events_processors = {
            ListenedEvents.ELECTRONIC_PAYMENT_CLOSE_DIALOGS: electronic_payment_close_dialog_processor,
            ListenedEvents.ELECTRONIC_PAYMENT_SHOW_QRCODE: electronic_payment_show_qrcode_processor,
            ListenedEvents.ELECTRONIC_PAYMENT_SHOW_CHOICES: electronic_payment_show_choices_processor,
            ListenedEvents.ELECTRONIC_PAYMENT_SHOW_INPUT: electronic_payment_show_input_processor,
            ListenedEvents.ELECTRONIC_PAYMENT_SHOW_MESSAGE: electronic_payment_show_message_processor,
            ListenedEvents.ELECTRONIC_PAYMENT_SHOW_QUESTION: electronic_payment_show_question_processor
        }

        return events_processors

    def _build_memory_cache(
        self,
    ):
        # type: (...) -> Optional[Dict[str, Any[Dict[str, CacheManager], CacheManager]]]

        if not self.flask_server:
            return None

        eft_time = self.config.refresh_cache_time.get("eft", 5)
        fiscal_time = self.config.refresh_cache_time.get("fiscal", 5)
        manager_time = self.config.refresh_cache_time.get("manager", 1)
        menu_time = self.config.refresh_cache_time.get("menu", 5)
        printer_time = self.config.refresh_cache_time.get("printer", 5)

        eft_cache_expiration = PeriodBasedExpiration(period_in_minutes=eft_time)
        fiscal_cache_expiration = PeriodBasedExpiration(period_in_minutes=fiscal_time)
        manager_cache_expiration = PeriodBasedExpiration(period_in_minutes=manager_time)
        menu_cache_expiration = PeriodBasedExpiration(period_in_minutes=menu_time)
        printer_cache_expiration = PeriodBasedExpiration(period_in_minutes=printer_time)

        fiscal_cache = CacheManager(
            cache_expiration=fiscal_cache_expiration,
            threaded=True,
        )
        menu_cache = CacheManager(
            cache_expiration=menu_cache_expiration,
            threaded=True,
        )

        pos_id_list = [str(pos_map.pos_id) for pos_map in self.config.api_key_mapping]

        eft_cache_manager = CacheManager(
            cache_expiration=eft_cache_expiration,
            threaded=True,
        )
        eft_cache = dict([
            (pos_id, eft_cache_manager)
            for pos_id in pos_id_list
        ])

        manager_cache_manager = CacheManager(
            cache_expiration=manager_cache_expiration,
            threaded=True,
        )
        manager_cache = dict([
            (pos_id, manager_cache_manager)
            for pos_id in pos_id_list
        ])

        printer_cache_manager = CacheManager(
            cache_expiration=printer_cache_expiration,
            threaded=True,
        )
        printer_cache = dict([
            (pos_id, printer_cache_manager)
            for pos_id in pos_id_list
        ])

        return dict(
            fiscal=fiscal_cache,
            menu=menu_cache,
            manager=manager_cache,
            printer=printer_cache,
            eft=eft_cache,
        )

    def _build_processors_routes(self):
        # type: () -> Dict[Route, MessageProcessor]

        customization_get_styles_interactor = GetStylesInteractor(
            logger=self.logger,
            config=self.config
        )
        order_add_items_interactor = \
            AddItemsInteractor(self.logger, self.config, self.pos_repository)
        order_cancel_interactor = \
            OrderCancelInteractor(self.logger, self.config, self.pos_repository)
        order_delete_line_interactor = \
            OrderDeleteLineInteractor(self.logger, self.config, self.pos_repository)
        order_close_shopping_cart_interactor = \
            CloseShoppingCartInteractor(self.logger, self.config, self.pos_repository, self.order_service)
        order_create_interactor = \
            OrderCreateInteractor(self.logger, self.config, self.pos_repository)
        order_add_properties_interactor = \
            AddPropertiesInteractor(self.logger, self.config, self.pos_repository)

        order_finalize_interactor = OrderFinalizeInteractor(
            self.logger,
            self.config,
            self.pos_repository,
            self.fiscal_service,
            self.printer_service,
        )

        order_get_detail_interactor = OrderGetDetailInteractor(
            self.logger,
            self.config,
            self.pos_repository,
            self.order_service,
        )

        order_insert_payment_interactor = OrderInsertPaymentInteractor(
            self.logger,
            self.config,
            self.pos_repository,
            self.fiscal_service,
            self.payment_service,
            self.electronic_payment_service,
        )

        order_change_items_quantity_interactor = \
            ChangeItemsQuantityInteractor(self.logger, self.config, self.pos_repository)
        order_reopen_shopping_cart_interactor = ReopenShoppingCartInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            order_service=self.order_service
        )
        order_apply_benefit_interactor = ApplyBenefitInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            order_service=self.order_service,
        )
        order_apply_coupon_interactor = ApplyCouponInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            coupon_service=self.coupon_service,
            order_service=self.order_service
        )
        order_remove_coupon_interactor = RemoveCouponInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            order_service=self.order_service
        )
        order_get_coupon_interactor = GetCouponInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            order_service=self.order_service,
        )
        menu_get_products_interactor = GetProductsInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            cache_menu_products=self.cache_menu_products,
        )
        menu_get_ruptured_products = GetRupturedProducts(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            cache_menu_ruptured_items=self.cache_menu_ruptured_items,
        )
        menu_get_navigation_interactor = GetNavigationInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            cache_navigation=self.cache_menu_navigation
        )
        menu_get_images_interactor = GetImagesInteractor(self.logger, self.config, self.images_service)
        menu_get_banners_interactor = GetBannersInteractor(
            logger=self.logger,
            config=self.config
        )
        menu_get_up_sell_part_codes_interactor = GetUpSellPartCodesInteractor(
            logger=self.logger,
            config=self.config
        )
        menu_get_product_info_interactor = GetProductInfoInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            product_info_service=self.product_info_service
        )
        payment_get_payment_types_interactor = GetPaymentTypesInteractor(
            self.logger,
            self.config,
            self.pos_repository,
            self.payment_service
        )
        peripheral_electronic_payment_read_messages_interactor = \
            ReadMessagesInteractor(
                self.logger,
                self.config,
                self.pos_repository,
                self.fiscal_service,
                self.electronic_messages
            )
        peripheral_electronic_payment_clear_messages_interactor = \
            ClearMessagesInteractor(
                self.logger,
                self.config,
                self.pos_repository,
                self.electronic_messages
            )
        peripheral_electronic_payment_send_responses_interactor = \
            SendResponsesInteractor(
                self.logger,
                self.config,
                self.pos_repository,
                self.fiscal_service,
                self.electronic_messages
            )
        peripheral_print_text_interactor = \
            PrintTextInteractor(self.logger, self.config, self.pos_repository)
        peripheral_read_scanner_interactor = \
            ReadScannerInteractor(self.logger, self.config, self.pos_repository)
        status_get_store_interactor = GetStoreInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            message_bus=self.message_bus,
            memory_cache=self.memory_cache,
        )
        cache_peripheral_electronic_payment_interactor = \
            CacheInteractor(self.logger, self.config, self.electronic_messages)

        customization_get_store_config_interactor = GetStoreConfigInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository
        )

        customization_get_custom_images_interactor = GetCustomImagesInteractor(
            logger=self.logger,
            config=self.config,
        )

        customization_get_text_interactor = GetTextInteractor(
            self.logger,
            self.config,
            self.pos_repository
        )

        order_cancel_payment_interactor = CancelPaymentInteractor(
            self.logger,
            self.config,
            self.pos_repository
        )

        order_save_interactor = OrderSaveInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            printer_service=self.printer_service,
        )

        search_customer_info_interactor = SearchCustomerInfoInteractor(
            logger=self.logger,
            config=self.config,
            pos_repository=self.pos_repository,
            loyalty_service=self.loyalty_service,
        )

        auto_mapped_apply_coupon_message_processor = PosLoggerAutoMappedMessageProcessor(
            request_type=ApplyCouponRequest,
            interactor=order_apply_coupon_interactor,
            exception_map={
                AuthValidationErrors: Responses.AuthValidation,
                ClientCustomError: Responses.ClientCustomError,
                ExternalServiceErrors: Responses.ExternalServicesError,
                InUseError: Responses.InUseError,
                NotAvailableError: Responses.NotAvailableError,
                NotFoundError: Responses.NotFoundError,
                OperationErrors: Responses.OperationErrors,
                ValidationError: Responses.ValidationError,
                BlockedError: Responses.BlockedError,
            },
            success_status=200,
            log_path=self.config.loader_path,
            component_name=self.config.component_name,
            logger=self.logger,
            logger_created=self.logger_created,
        )

        processor_routes = {
            Route("GET", "/order/getDetail/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=OrderGetDetailRequest,
                interactor=order_get_detail_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/create/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=OrderCreateRequest,
                interactor=order_create_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=201,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/addProperties/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=AddPropertiesRequest,
                interactor=order_add_properties_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/addItems/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=AddItemsRequest,
                interactor=order_add_items_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/changeItemsQuantity/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=ChangeItemsQuantityRequest,
                interactor=order_change_items_quantity_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/insertPayment/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=OrderInsertPaymentRequest,
                interactor=order_insert_payment_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors,
                    BadRequestError: Responses.BadRequestError,
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/registerPayment/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=OrderInsertPaymentRequest,
                interactor=order_insert_payment_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors,
                    BadRequestError: Responses.BadRequestError,
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("DELETE", "/order/cancelPayment/{posId}/{tenderId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=CancelPaymentRequest,
                interactor=order_cancel_payment_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/cancel/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=OrderCancelRequest,
                interactor=order_cancel_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("DELETE", "/order/deleteLine/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=OrderDeleteLineRequest,
                interactor=order_delete_line_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/closeShoppingCart/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=CloseShoppingCartRequest,
                interactor=order_close_shopping_cart_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/reopenShoppingCart/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=ReopenShoppingCartRequest,
                interactor=order_reopen_shopping_cart_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/finalize/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=OrderFinalizeRequest,
                interactor=order_finalize_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/applyCoupon/{posId}"): auto_mapped_apply_coupon_message_processor,
            Route("POST", "/order/applyCoupon/{posId}/{language}"): auto_mapped_apply_coupon_message_processor,

            Route("DELETE", "/order/removeCoupon/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=RemoveCouponRequest,
                interactor=order_remove_coupon_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),
            Route("POST", "/order/save/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=OrderSaveRequest,
                interactor=order_save_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors,
                    BadRequestError: Responses.BadRequestError,
                },
                success_status=200,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/order/getCoupon/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetCouponRequest,
                interactor=order_get_coupon_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=200,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/order/applyBenefit/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=ApplyBenefitRequest,
                interactor=order_apply_benefit_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors,
                },
                success_status=200,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/menu/products"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetProductsRequest,
                interactor=menu_get_products_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                response_content_type="application/gzip",
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/menu/rupturedProducts"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetRupturedProductsRequest,
                interactor=menu_get_ruptured_products,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/menu/navigation/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetNavigationRequest,
                interactor=menu_get_navigation_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/menu/productInfo/{language}"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetProductInfoRequest,
                interactor=menu_get_product_info_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/menu/images"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetImagesRequest,
                interactor=menu_get_images_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/menu/banners"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetBannersRequest,
                interactor=menu_get_banners_interactor,
                exception_map={
                    AuthValidationErrors: Responses.AuthValidation
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/menu/upSellPartCodes"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetUpSellPartCodesRequest,
                interactor=menu_get_up_sell_part_codes_interactor,
                exception_map={
                    AuthValidationErrors: Responses.AuthValidation
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/payment/paymentTypes"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetPaymentTypesRequest,
                interactor=payment_get_payment_types_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/peripheral/electronicPayment/readMessages/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=ReadMessagesRequest,
                interactor=peripheral_electronic_payment_read_messages_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/peripheral/electronicPayment/clearMessages/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=ClearMessagesRequest,
                interactor=peripheral_electronic_payment_clear_messages_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/peripheral/electronicPayment/sendResponse/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=SendResponsesRequest,
                interactor=peripheral_electronic_payment_send_responses_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=204,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/peripheral/printText/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=PrintTextRequest,
                interactor=peripheral_print_text_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError
                },
                success_status=204,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/peripheral/readScanner/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=ReadScannerRequest,
                interactor=peripheral_read_scanner_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                success_status=200,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/status/store/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetStoreRequest,
                interactor=status_get_store_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("POST", "/cache/peripheral/electronicPayment"): PosLoggerAutoMappedMessageProcessor(
                request_type=CacheRequest,
                interactor=cache_peripheral_electronic_payment_interactor,
                exception_map={},
                success_status=204,
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/customization/getStyles"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetStylesRequest,
                interactor=customization_get_styles_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/customization/getText"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetTextRequest,
                interactor=customization_get_text_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/customization/storeConfig/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetStoreConfigRequest,
                interactor=customization_get_store_config_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/customization/getCustomImages"): PosLoggerAutoMappedMessageProcessor(
                request_type=GetCustomImagesRequest,
                interactor=customization_get_custom_images_interactor,
                exception_map={
                    AuthValidationErrors: Responses.AuthValidation,
                    OperationErrors: Responses.OperationErrors,
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),

            Route("GET", "/loyalty/getCustomer/{posId}"): PosLoggerAutoMappedMessageProcessor(
                request_type=SearchCustomerInfoRequest,
                interactor=search_customer_info_interactor,
                exception_map={
                    ValidationError: Responses.ValidationError,
                    AuthValidationErrors: Responses.AuthValidation,
                    ExternalServiceErrors: Responses.ExternalServicesError,
                    OperationErrors: Responses.OperationErrors,
                    NotFoundError: Responses.NotFoundError,
                },
                log_path=self.config.loader_path,
                component_name=self.config.component_name,
                logger=self.logger,
                logger_created=self.logger_created,
            ),
        }

        return processor_routes



from _CashTransferSummaryInfo import CashTransferSummaryInfo
from _CashTransferSummaryDto import CashTransferSummaryDto
from _CashTransferSummaryHeaderDto import CashTransferSummaryHeaderDto
from _CashTransferSummaryBodyDto import CashTransferSummaryBodyDto


__all__ = ["CashTransferSummaryInfo",
           "CashTransferSummaryDto",
           "CashTransferSummaryHeaderDto",
           "CashTransferSummaryBodyDto"]



from _StatusResponse import StatusResponse


class GetStoreResponse(object):
    def __init__(
        self,
        menu_id,        # type: str
        menu_version,   # type: str
        menu_name,      # type: str
        printer,        # type: StatusResponse
        eft,            # type: StatusResponse
        fiscal,         # type: StatusResponse
        manager,        # type: StatusResponse
    ):
        # type: (...) -> None

        self.menu_id = menu_id
        self.menu_version = menu_version
        self.menu_name = menu_name
        self.printer = printer
        self.eft = eft
        self.fiscal = fiscal
        self.manager = manager



import logging

from logging import Logger
from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
    Type,
    Union,
)

from helper import config_logger
from httpmessageprocessorutil.automapped import AutoMappedMessageProcessor
from httpmessageprocessor import ApiRequest
from messagebus import (
    Event,
    Message,
    MessageBus,
)


class PosLoggerAutoMappedMessageProcessor(AutoMappedMessageProcessor):

    def __init__(
        self,
        request_type,                   # type: Optional[Type]
        interactor,                     # type: Any
        logger_created,                 # type: dict
        exception_map=None,             # type: Optional[Dict[type, Union[int, Tuple[int, str]]]]
        success_status=200,             # type: int
        extra_mapping=None,             # type: dict
        event_name=None,                # type: Optional[str]
        response_content_type=None,     # type: Optional[str]
        log_path=None,                  # type: Optional[str]
        component_name=None,            # type: Optional[str]
        logger=None,                    # type: Optional[Logger]
    ):
        # type: (...) -> None

        super(PosLoggerAutoMappedMessageProcessor, self).__init__(
            request_type=request_type,
            interactor=interactor,
            exception_map=exception_map,
            success_status=success_status,
            extra_mapping=extra_mapping,
            event_name=event_name,
            response_content_type=response_content_type,
        )

        self.logger_created = logger_created
        self.logger = logger
        self.component_name = component_name
        self.log_path = log_path
        self.interactor = interactor

    def parse_data(
        self,
        api_request,    # type: ApiRequest
        event_type,     # type: Optional[str]
    ):
        # type: (...) -> Any

        try:
            pos_id = api_request.path_parameters.get("posId")
            if pos_id:
                log_name = "{}_{:02d}".format(
                    self.component_name,
                    int(pos_id),
                )
                if log_name not in self.logger_created:
                    config_logger(
                        loader_path=self.log_path,
                        log_name=log_name,
                        root_level=logging.DEBUG,
                    )
                    self.logger_created[log_name] = True

                internal_order_logger = logging.getLogger(name=log_name)
                self.interactor.logger = internal_order_logger
                self.interactor.logger.info(msg=api_request)
        except Exception as e:
            if self.logger:
                self.logger.exception(e)

        return super(PosLoggerAutoMappedMessageProcessor, self).parse_data(
            api_request=api_request,
            event_type=event_type,
        )

    def format_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        input_obj,      # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        self.interactor.logger.exception(exception)
        super(PosLoggerAutoMappedMessageProcessor, self).format_exception(
            message_bus=message_bus,
            message=message,
            event=event,
            input_obj=input_obj,
            exception=exception,
        )

    def format_parse_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        data,           # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        self.interactor.logger.exception(exception)
        super(PosLoggerAutoMappedMessageProcessor, self).format_parse_exception(
            message_bus=message_bus,
            message=message,
            event=event,
            data=data,
            exception=exception,
        )



class ConfigurationReaderException(Exception):
    pass


class InvalidMessageContentException(Exception):
    pass


class OrderNotFoundException(Exception):
    def __init__(self):
        super(Exception, self).__init__("NDISCOUNT_ORDER_NOT_FOUND")


class BenefitNotFound(Exception):
    pass


class TimeoutException(Exception):
    def __init__(self):
        super(Exception, self).__init__("NDISCOUNT_TIMEOUT")


class DiscountNotFound(Exception):
    def __init__(self):
        super(Exception, self).__init__("DISCOUNT_NOT_FOUND")


class CannotRemoveItemsInATotaledOrder(Exception):
    def __init__(self):
        super(Exception, self).__init__("CANNOT_REMOVE_ITEMS_IN_A_TOTALED_ORDER")


class ProductNotFound(Exception):
    def __init__(self):
        super(Exception, self).__init__("NDISCOUNT_PRODUCT_NOT_FOUND")


class InvalidTenderIdException(Exception):
    def __init__(self):
        super(Exception, self).__init__("NDISCOUNT_INVALID_TENDER_ID")


class InvalidAmountException(Exception):
    def __init__(self):
        super(Exception, self).__init__("NDISCOUNT_INVALID_AMOUNT")


class InvalidPercentageException(Exception):
    def __init__(self):
        super(Exception, self).__init__("NDISCOUNT_INVALID_PERCENTAGE")


class NotEffectiveDiscountException(Exception):
    def __init__(self):
        super(Exception, self).__init__("NDISCOUNT_NOT_EFFECTIVE_DISCOUNT")


class InvalidOperationException(Exception):
    def __init__(self):
        super(Exception, self).__init__("NDISCOUNT_INVALID_OPERATION")


class InvalidQuantityException(Exception):
    def __init__(self):
        super(Exception, self).__init__()


class AddItemsOnTotalException(Exception):
    def __init__(self):
        super(Exception, self).__init__("NDISCOUNT_CANNOT_ADD_ITEMS_ON_TOTALED_ORDER")


class CircularDependencyToApplyDiscountException(Exception):
    def __init__(
        self,
    ):
        super(Exception, self).__init__("NDISCOUNT_CIRCULAR_DEPENDENCY_TRYING_TO_APPLY_DISCOUNT")
